# -*- coding: utf-8 -*-
"""
sup3r data collection CLI entry points.
"""
import click
import logging

from reV.pipeline.status import Status

from rex.utilities.execution import SubprocessManager
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_mult
from rex.utilities.utilities import safe_json_load

from sup3r.utilities import ModuleName
from sup3r.version import __version__
from sup3r.postprocessing.collection import Collector


logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, verbose):
    """Sup3r Data Collection Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='sup3r data collection configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run sup3r data collection from a config file.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    config_file : str
        Filepath to sup3r data collection json file.
    verbose : bool
        Flag to turn on debug logging. Default is not verbose.
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    config = safe_json_load(config_file)
    config_verbose = config.get('log_level', 'INFO')
    config_verbose = (config_verbose == 'DEBUG')
    verbose = any([verbose, config_verbose, ctx.obj['VERBOSE']])

    init_mult('sup3r_data_collect', './logs/', modules=[__name__, 'sup3r'],
              verbose=verbose)

    exec_kwargs = config.get('execution_control', {})
    hardware_option = exec_kwargs.get('option', 'local')
    logger.debug('Found execution kwargs: {}'.format(exec_kwargs))
    logger.debug('Hardware run option: "{}"'.format(hardware_option))

    name = 'sup3r_data_collect'
    ctx.obj['NAME'] = name

    cmd = Collector.get_node_cmd(config)
    logger.debug(f'Running command: {cmd}')

    if hardware_option.lower() in ('eagle', 'slurm'):
        kickoff_slurm_job(ctx, cmd, **exec_kwargs)
    else:
        SubprocessManager.submit(cmd)


def kickoff_slurm_job(ctx, cmd, alloc='sup3r', memory=None, walltime=4,
                      feature=None, stdout_path='./stdout/'):
    """Run sup3r on HPC via SLURM job submission.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    cmd : str
        Command to be submitted in SLURM shell script. Example:
            'python -m sup3r.cli data_extract -c <config_file>'
    alloc : str
        HPC project (allocation) handle. Example: 'sup3r'. Default is not to
        state an allocation (does not work on Eagle slurm).
    memory : int
        Node memory request in GB.
    walltime : float
        Node walltime request in hours. Default is not to state a walltime
        (does not work on Eagle slurm).
    feature : str
        Additional flags for SLURM job. Format is "--qos=high"
        or "--depend=[state:job_id]". Default is None.
    stdout_path : str
        Path to print .stdout and .stderr files.
    """

    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']
    fn_out = ctx.obj['FN_OUT']

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir,
                                        module=ModuleName.DATA_COLLECT,
                                        job_name=name,
                                        hardware='slurm',
                                        subprocess_manager=slurm_manager)

    msg = 'sup3r data collection CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'.format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info(
            'Running sup3r data collection on SLURM with node name "{}".'
            .format(name))

        out = slurm_manager.sbatch(cmd,
                                   alloc=alloc,
                                   memory=memory,
                                   walltime=walltime,
                                   feature=feature,
                                   name=name,
                                   stdout_path=stdout_path)[0]
        if out:
            msg = ('Kicked off sup3r data collection job "{}" '
                   '(SLURM jobid #{}).'
                   .format(name, out))

        # add job to sup3r status file.
        Status.add_job(out_dir, module=ModuleName.DATA_COLLECT,
                       job_name=name, replace=True,
                       job_attrs={'job_id': out, 'hardware': 'slurm',
                                  'fn_out': fn_out, 'out_dir': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running sup3r data collection CLI')
        raise



# -*- coding: utf-8 -*-
"""Sup3r model software"""
import copy
import os
import time
import json
import logging
import numpy as np
import pprint
import pandas as pd
import tensorflow as tf
from tensorflow.keras import optimizers
from rex.utilities.utilities import safe_json_load
from phygnn import CustomNetwork
from warnings import warn

from sup3r.utilities import VERSION_RECORD
import sup3r.utilities.loss_metrics


logger = logging.getLogger(__name__)


class Sup3rGan:
    """Basic sup3r GAN model."""

    def __init__(self, gen_layers, disc_layers, loss='MeanSquaredError',
                 optimizer=None, learning_rate=1e-4,
                 optimizer_disc=None, learning_rate_disc=None,
                 history=None, meta=None,
                 means=None, stdevs=None, name=None):
        """
        Parameters
        ----------
        gen_layers : list | str
            Hidden layers input argument to phygnn.base.CustomNetwork for the
            generative super resolving model. Can also be a str filepath to a
            .json config file containing the input layers argument or a .pkl
            for a saved pre-trained model.
        disc_layers : list | str
            Hidden layers input argument to phygnn.base.CustomNetwork for the
            discriminative model (spatial or spatiotemporal discriminator). Can
            also be a str filepath to a .json config file containing the input
            layers argument or a .pkl for a saved pre-trained model.
        loss : str
            Loss function class name from sup3r.utilities.loss_metrics
            (prioritized) or tensorflow.keras.losses. Defaults to
            tf.keras.losses.MeanSquaredError.
        optimizer : tf.keras.optimizers.Optimizer | dict | None | str
            Instantiated tf.keras.optimizers object or a dict optimizer config
            from tf.keras.optimizers.get_config(). None defaults to Adam.
        learning_rate : float, optional
            Optimizer learning rate. Not used if optimizer input arg is a
            pre-initialized object or if optimizer input arg is a config dict.
        optimizer_disc : tf.keras.optimizers.Optimizer | dict | None
            Same as optimizer input, but if specified this makes a different
            optimizer just for the discriminator network (spatial or
            spatiotemporal disc).
        learning_rate_disc : float, optional
            Same as learning_rate input, but if specified this makes a
            different learning_rate just for the discriminator network (spatial
            or spatiotemporal disc).
        history : pd.DataFrame | str | None
            Model training history with "epoch" index, str pointing to a saved
            history csv file with "epoch" as first column, or None for clean
            history
        meta : dict | None
            Model meta data that describes how the model was created.
        means : np.ndarray | list | None
            Set of mean values for data normalization with the same length as
            number of features. Can be used to maintain a consistent
            normalization scheme between transfer learning domains.
        stdevs : np.ndarray | list | None
            Set of stdev values for data normalization with the same length as
            number of features. Can be used to maintain a consistent
            normalization scheme between transfer learning domains.
        name : str | None
            Optional name for the GAN.
        """

        self._version_record = VERSION_RECORD
        logger.info('Active python environment versions: \n{}'
                    .format(pprint.pformat(self._version_record, indent=4)))

        self.name = name if name is not None else self.__class__.__name__
        self._meta = meta if meta is not None else {}

        self.loss_name = loss
        self.loss_fun = self.get_loss_fun(loss)

        self._history = history
        if isinstance(self._history, str):
            self._history = pd.read_csv(self._history, index_col=0)

        optimizer_disc = optimizer_disc or copy.deepcopy(optimizer)
        learning_rate_disc = learning_rate_disc or learning_rate
        self._optimizer = self.init_optimizer(optimizer, learning_rate)
        self._optimizer_disc = self.init_optimizer(optimizer_disc,
                                                   learning_rate_disc)

        self._gen = self.load_network(gen_layers, 'generator')
        self._disc = self.load_network(disc_layers, 'discriminator')

        self._means = (means if means is None
                       else np.array(means).astype(np.float32))
        self._stdevs = (stdevs if stdevs is None
                        else np.array(stdevs).astype(np.float32))

    @staticmethod
    def init_optimizer(optimizer, learning_rate):
        """Initialize keras optimizer object.

        Parameters
        ----------
        optimizer : tf.keras.optimizers.Optimizer | dict | None | str
            Instantiated tf.keras.optimizers object or a dict optimizer config
            from tf.keras.optimizers.get_config(). None defaults to Adam.
        learning_rate : float, optional
            Optimizer learning rate. Not used if optimizer input arg is a
            pre-initialized object or if optimizer input arg is a config dict.

        Returns
        -------
        optimizer : tf.keras.optimizers.Optimizer
            Initialized optimizer object.
        """
        if isinstance(optimizer, dict):
            class_name = optimizer['name']
            OptimizerClass = getattr(optimizers, class_name)
            optimizer = OptimizerClass.from_config(optimizer)
        elif optimizer is None:
            optimizer = optimizers.Adam(learning_rate=learning_rate)

        return optimizer

    def load_network(self, model, name):
        """Load a CustomNetwork object from hidden layers config, .json file
        config, or .pkl file saved pre=trained model.

        Parameters
        ----------
        model : str | dict
            Model hidden layers config, a .json with "hidden_layers" key, or a
            .pkl for a saved pre-trained model.
        name : str
            Name of the model to be loaded

        Returns
        -------
        model : phygnn.CustomNetwork
            CustomNetwork object initialized from the model input.
        """

        if isinstance(model, str) and model.endswith('.json'):
            model = safe_json_load(model)
            self._meta[f'config_{name}'] = model
            if 'hidden_layers' in model:
                model = model['hidden_layers']
            else:
                msg = ('Could not load model from json config, need '
                       '"hidden_layers" key at top level but only found: {}'
                       .format(model.keys()))
                logger.error(msg)
                raise KeyError(msg)

        elif isinstance(model, str) and model.endswith('.pkl'):
            model = CustomNetwork.load(model)

        if isinstance(model, list):
            model = CustomNetwork(hidden_layers=model, name=name)

        if not isinstance(model, CustomNetwork):
            msg = ('Something went wrong. Tried to load a custom network '
                   'but ended up with a model of type "{}"'
                   .format(type(model)))
            logger.error(msg)
            raise TypeError(msg)

        return model

    def save(self, out_dir):
        """Save the GAN with its sub-networks to a directory.

        Parameters
        ----------
        out_dir : str
            Directory to save GAN model files. This directory will be created
            if it does not already exist.
        """

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        fp_gen = os.path.join(out_dir, 'model_gen.pkl')
        self.generator.save(fp_gen)

        fp_disc = os.path.join(out_dir, 'model_disc.pkl')
        self.discriminator.save(fp_disc)

        fp_history = None
        if isinstance(self.history, pd.DataFrame):
            fp_history = os.path.join(out_dir, 'history.csv')
            self.history.to_csv(fp_history)

        fp_params = os.path.join(out_dir, 'model_params.json')
        with open(fp_params, 'w') as f:
            params = self.model_params
            json.dump(params, f, sort_keys=True, indent=2)

        logger.info('Saved GAN to disk in directory: {}'.format(out_dir))

    @classmethod
    def load(cls, out_dir):
        """Load the GAN with its sub-networks from a previously saved-to output
        directory.

        Parameters
        ----------
        out_dir : str
            Directory to load GAN model files from.

        Returns
        -------
        out : BaseModel
            Returns a pretrained gan model that was previously saved to out_dir
        """

        logger.info('Loading GAN from disk in directory: {}'.format(out_dir))

        fp_gen = os.path.join(out_dir, 'model_gen.pkl')
        fp_disc = os.path.join(out_dir, 'model_disc.pkl')
        params = cls._load_saved_params(out_dir)

        return cls(fp_gen, fp_disc, **params)

    @staticmethod
    def _load_saved_params(out_dir):
        """Load saved GAN model_params (you need this and the gen+disc models
        to load a full GAN).

        Parameters
        ----------
        out_dir : str
            Directory to load GAN model files from.

        Returns
        -------
        params : dict
            Model parameters loaded from disk json file. This should be the
            same as self.model_params with and additional 'history' entry.
            Should be all the kwargs you need to init a GAN model.
        """

        fp_params = os.path.join(out_dir, 'model_params.json')
        with open(fp_params, 'r') as f:
            params = json.load(f)

        # using the saved model dir makes this more portable
        fp_history = os.path.join(out_dir, 'history.csv')
        if os.path.exists(fp_history):
            params['history'] = fp_history
        else:
            params['history'] = None

        if 'version_record' in params:
            version_record = params.pop('version_record')
            logger.info('Loading GAN from disk that was created with the '
                        'following package versions: \n{}'
                        .format(pprint.pformat(version_record, indent=2)))

        return params

    @property
    def means(self):
        """Get the data normalization mean values.

        Returns
        -------
        np.ndarray
        """
        return self._means

    @property
    def stdevs(self):
        """Get the data normalization standard deviation values.

        Returns
        -------
        np.ndarray
        """
        return self._stdevs

    def set_norm_stats(self, new_means, new_stdevs):
        """Set the normalization statistics associated with a data batch
        handler to model attributes.

        Parameters
        ----------
        new_means : list | tuple | np.ndarray
            1D iterable of mean values with same length as number of features.
        new_stdevs : list | tuple | np.ndarray
            1D iterable of stdev values with same length as number of features.
        """

        if self._means is not None:
            logger.info('Setting new normalization statistics...')
            logger.info("Model's previous data mean values: {}"
                        .format(self._means))
            logger.info("Model's previous data stdev values: {}"
                        .format(self._stdevs))

        self._means = new_means
        self._stdevs = new_stdevs

        if not isinstance(self._means, np.ndarray):
            self._means = np.array(self._means)
        if not isinstance(self._stdevs, np.ndarray):
            self._stdevs = np.array(self._stdevs)

        logger.info('Set data normalization mean values: {}'
                    .format(self._means))
        logger.info('Set data normalization stdev values: {}'
                    .format(self._stdevs))

    def set_feature_names(self, training_features, output_features):
        """Set the list of feature names input/output to/from the generative
        model and the discriminator(s)

        Parameters
        ----------
        training_features : list
            List of features names on the input side of the GAN
        output_features : list
            List of feature names output by the GAN
        """

        if self.training_features is None:
            self.meta['training_features'] = training_features
        elif self.training_features != training_features:
            msg = ('GAN was previously trained on feature list {} but '
                   'received new features {}'
                   .format(self.training_features, training_features))
            logger.error(msg)
            raise KeyError(msg)

        if self.output_features is None:
            self.meta['output_features'] = output_features
        elif self.output_features != output_features:
            msg = ('GAN was previously trained to output features {} but '
                   'received data to output new features {}'
                   .format(self.output_features, output_features))
            logger.error(msg)
            raise KeyError(msg)

    @staticmethod
    def get_loss_fun(loss):
        """Get the initialized loss function class from the sup3r loss library
        or the tensorflow losses.

        Parameters
        ----------
        loss : str
            Loss function class name from sup3r.utilities.loss_metrics
            (prioritized) or tensorflow.keras.losses. Defaults to
            tf.keras.losses.MeanSquaredError.

        Returns
        -------
        out : tf.keras.losses.Loss
            Initialized loss function class that is callable, e.g. if
            "MeanSquaredError" is requested, this will return
            an instance of tf.keras.losses.MeanSquaredError()
        """

        out = getattr(sup3r.utilities.loss_metrics, loss, None)
        if out is None:
            out = getattr(tf.keras.losses, loss, None)

        if out is None:
            msg = ('Could not find requested loss function "{}" in '
                   'sup3r.utilities.loss_metrics or tf.keras.losses.'
                   .format(loss))
            logger.error(msg)
            raise KeyError(msg)

        return out()

    @staticmethod
    def seed(s=0):
        """
        Set the random seed for reproducible results.

        Parameters
        ----------
        s : int
            Random seed
        """
        CustomNetwork.seed(s=s)

    @property
    def generator(self):
        """Get the generative model.

        Returns
        -------
        phygnn.base.CustomNetwork
        """
        return self._gen

    @property
    def generator_weights(self):
        """Get a list of layer weights and bias terms for the generator model.

        Returns
        -------
        list
        """
        return self.generator.weights

    def generate(self, low_res, norm_in=True, un_norm_out=True):
        """Use the generator model to generate high res data from low res
        input. This is the public generate function.

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution input data, usually a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        norm_in : bool
            Flag to normalize low_res input data if the self._means,
            self._stdevs attributes are available. The generator should always
            received normalized data with mean=0 stdev=1.
        un_norm_out : bool
           Flag to un-normalize synthetically generated output data to physical
           units

        Returns
        -------
        hi_res : ndarray
            Synthetically generated high-resolution data, usually a 4D or 5D
            array with shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        """

        if norm_in and self._means is not None:
            low_res = low_res.copy()
            for i, (m, s) in enumerate(zip(self._means, self._stdevs)):
                low_res[..., i] -= m
                if s > 0:
                    low_res[..., i] /= s
                else:
                    msg = ('Standard deviation is zero for '
                           f'{self.training_features[i]}')
                    logger.warning(msg)
                    warn(msg)

        hi_res = self.generator.layers[0](low_res)
        for i, layer in enumerate(self.generator.layers[1:]):
            try:
                hi_res = layer(hi_res)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, hi_res.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        hi_res = hi_res.numpy()

        if un_norm_out and self._means is not None:
            for i, feature in enumerate(self.training_features):
                if feature in self.output_features:
                    m = self._means[i]
                    s = self._stdevs[i]
                    j = self.output_features.index(feature)
                    hi_res[..., j] = (hi_res[..., j] * s) + m

        return hi_res

    @tf.function
    def _tf_generate(self, low_res):
        """Use the generator model to generate high res data from los res input

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data. The generator should always
            received normalized data with mean=0 stdev=1.

        Returns
        -------
        hi_res : tf.Tensor
            Synthetically generated high-resolution data
        """

        hi_res = self.generator.layers[0](low_res)
        for i, layer in enumerate(self.generator.layers[1:]):
            try:
                hi_res = layer(hi_res)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, hi_res.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        return hi_res

    @property
    def discriminator(self):
        """Get the discriminator model.

        Returns
        -------
        phygnn.base.CustomNetwork
        """
        return self._disc

    @property
    def discriminator_weights(self):
        """Get a list of layer weights and bias terms for the discriminator
        model.

        Returns
        -------
        list
        """
        return self.discriminator.weights

    def discriminate(self, hi_res, norm_in=False):
        """Run the discriminator model on a hi resolution input field.

        Parameters
        ----------
        hi_res : np.ndarray
            Real or fake high res data in a 4D or 5D tensor:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        norm_in : bool
            Flag to normalize low_res input data if the self._means,
            self._stdevs attributes are available. The disc should always
            received normalized data with mean=0 stdev=1.

        Returns
        -------
        out : np.ndarray
            Discriminator output logits
        """

        if isinstance(hi_res, tf.Tensor):
            hi_res = hi_res.numpy()

        if norm_in and self._means is not None:
            hi_res = hi_res if isinstance(hi_res, tf.Tensor) else hi_res.copy()
            for i, (m, s) in enumerate(zip(self._means, self._stdevs)):
                islice = tuple([slice(None)] * (len(hi_res.shape) - 1) + [i])
                hi_res[islice] = (hi_res[islice] - m) / s

        out = self.discriminator.layers[0](hi_res)
        for i, layer in enumerate(self.discriminator.layers[1:]):
            try:
                out = layer(out)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, out.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        out = out.numpy()

        return out

    @tf.function
    def _tf_discriminate(self, hi_res):
        """Run the discriminator model on a hi resolution input field.

        Parameters
        ----------
        hi_res : np.ndarray
            Real or fake high res data in a 4D or 5D tensor:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
            This input should always be normalized with mean=0 and stdev=1

        Returns
        -------
        out : np.ndarray
            Discriminator output logits
        """

        out = self.discriminator.layers[0](hi_res)
        for i, layer in enumerate(self.discriminator.layers[1:]):
            try:
                out = layer(out)
            except Exception as e:
                msg = ('Could not run layer #{} "{}" on tensor of shape {}'
                       .format(i + 1, layer, out.shape))
                logger.error(msg)
                raise RuntimeError(msg) from e

        return out

    def un_norm_output(self, hi_res):
        """Un-normalize synthetically generated output data to physical units

        Parameters
        ----------
        hi_res : tf.Tensor | np.ndarray
            Synthetically generated high-resolution data

        Returns
        -------
        hi_res : np.ndarray
            Synthetically generated high-resolution data
        """
        if self._means is not None:
            if isinstance(hi_res, tf.Tensor):
                hi_res = hi_res.numpy()

            for i, feature in enumerate(self.training_features):
                if feature in self.output_features:
                    m = self._means[i]
                    s = self._stdevs[i]
                    j = self.output_features.index(feature)
                    hi_res[..., j] = (hi_res[..., j] * s) + m

        return hi_res

    @property
    def version_record(self):
        """A record of important versions that this model was built with.

        Returns
        -------
        dict
        """
        return self._version_record

    @property
    def optimizer(self):
        """Get the tensorflow optimizer to perform gradient descent
        calculations for the generative network. This is functionally identical
        to optimizer_disc is no special optimizer model or learning rate was
        specified for the disc.

        Returns
        -------
        tf.keras.optimizers.Optimizer
        """
        return self._optimizer

    @property
    def optimizer_disc(self):
        """Get the tensorflow optimizer to perform gradient descent
        calculations for the discriminator network.

        Returns
        -------
        tf.keras.optimizers.Optimizer
        """
        return self._optimizer_disc

    @staticmethod
    def get_optimizer_config(optimizer):
        """Get a config that defines the current GAN optimizer

        Parameters
        ----------
        optimizer : tf.keras.optimizers.Optimizer
            TF-Keras optimizer object

        Returns
        -------
        config : dict
            Optimizer config
        """
        conf = optimizer.get_config()
        for k, v in conf.items():
            # need to convert numpy dtypes to float/int for json.dump()
            if np.issubdtype(type(v), np.floating):
                conf[k] = float(v)
            elif np.issubdtype(type(v), np.integer):
                conf[k] = int(v)
        return conf

    def update_optimizer(self, option='generator', **kwargs):
        """Update optimizer by changing current configuration

        Parameters
        ----------
        option : str
            Which optimizer to update. Can be "generator", "discriminator", or
            "all"
        kwargs : dict
            kwargs to use for optimizer configuration update
        """

        if 'gen' in option.lower() or 'all' in option.lower():
            conf = self.get_optimizer_config(self.optimizer)
            conf.update(**kwargs)
            OptimizerClass = getattr(optimizers, conf['name'])
            self._optimizer = OptimizerClass.from_config(conf)

        if 'disc' in option.lower() or 'all' in option.lower():
            conf = self.get_optimizer_config(self.optimizer_disc)
            conf.update(**kwargs)
            OptimizerClass = getattr(optimizers, conf['name'])
            self._optimizer_disc = OptimizerClass.from_config(conf)

    @property
    def meta(self):
        """Get meta data dictionary that defines how the model was created"""

        if 'class' not in self._meta:
            self._meta['class'] = self.__class__.__name__

        return self._meta

    @property
    def training_features(self):
        """Get the list of input feature names that the generative model was
        trained on."""
        return self.meta.get('training_features', None)

    @property
    def output_features(self):
        """Get the list of output feature names that the generative model
        outputs and that the discriminator predicts on."""
        return self.meta.get('output_features', None)

    @property
    def model_params(self):
        """
        Model parameters, used to save model to disc

        Returns
        -------
        dict
        """
        means = (self._means if self._means is None
                 else [float(m) for m in self._means])
        stdevs = (self._stdevs if self._stdevs is None
                  else [float(s) for s in self._stdevs])

        config_optm_g = self.get_optimizer_config(self.optimizer)
        config_optm_d = self.get_optimizer_config(self.optimizer_disc)

        model_params = {'name': self.name,
                        'loss': self.loss_name,
                        'version_record': self.version_record,
                        'optimizer': config_optm_g,
                        'optimizer_disc': config_optm_d,
                        'means': means,
                        'stdevs': stdevs,
                        'meta': self.meta,
                        }

        return model_params

    @property
    def history(self):
        """
        Model training history DataFrame (None if not yet trained)

        Returns
        -------
        pandas.DataFrame | None
        """
        return self._history

    @property
    def weights(self):
        """Get a list of all the layer weights and bias terms for the
        generator and discriminator networks
        """
        return self.generator_weights + self.discriminator_weights

    @staticmethod
    def get_weight_update_fraction(history, comparison_key,
                                   update_bounds=(0.5, 0.95),
                                   update_frac=0.0):
        """Get the factor by which to multiply previous adversarial loss
        weight

        Parameters
        ----------
        history : dict
            Dictionary with information on how often discriminators
            were trained during previous epoch.
        comparison_key : str
            history key to use for update check
        update_bounds : tuple
            Tuple specifying allowed range for history[comparison_key]. If
            history[comparison_key] < update_bounds[0] then the weight will be
            increased by (1 + update_frac). If history[comparison_key] >
            update_bounds[1] then the weight will be decreased by 1 / (1 +
            update_frac).
        update_frac : float
            Fraction by which to increase/decrease adversarial loss weight

        Returns
        -------
        float
            Factor by which to multiply old weight to get updated weight
        """

        val = history[comparison_key]
        if isinstance(val, (list, tuple, np.ndarray)):
            val = val[-1]

        if val < update_bounds[0]:
            return 1 + update_frac
        elif val > update_bounds[1]:
            return 1 / (1 + update_frac)
        else:
            return 1

    @staticmethod
    def update_loss_details(loss_details, new_data, batch_len, prefix=None):
        """Update a dictionary of loss_details with loss information from a new
        batch.

        Parameters
        ----------
        loss_details : dict
            Namespace of the breakdown of loss components where each value is a
            running average at the current state in the epoch.
        new_data : dict
            Namespace of the breakdown of loss components for a single new
            batch.
        batch_len : int
            Length of the incomming batch.
        prefix : None | str
            Option to prefix the names of the loss data when saving to the
            loss_details dictionary.

        Returns
        -------
        loss_details : dict
            Same as input loss_details but with running averages updated.
        """
        assert 'n_obs' in loss_details, 'loss_details must have n_obs to start'
        prior_n_obs = loss_details['n_obs']
        new_n_obs = prior_n_obs + batch_len

        for key, new_value in new_data.items():
            key = key if prefix is None else prefix + key
            new_value = (new_value if not isinstance(new_value, tf.Tensor)
                         else new_value.numpy())

            if key in loss_details:
                saved_value = loss_details[key]
                saved_value *= prior_n_obs
                saved_value += batch_len * new_value
                saved_value /= new_n_obs
                loss_details[key] = saved_value
            else:
                loss_details[key] = new_value

        loss_details['n_obs'] = new_n_obs

        return loss_details

    @staticmethod
    def log_loss_details(loss_details, level='INFO'):
        """Log the loss details to the module logger.

        Parameters
        ----------
        loss_details : dict
            Namespace of the breakdown of loss components where each value is a
            running average at the current state in the epoch.
        level : str
            Log level (e.g. INFO, DEBUG)
        """
        for k, v in sorted(loss_details.items()):
            if k != 'n_obs':
                if isinstance(v, str):
                    msg_format = '\t{}: {}'
                else:
                    msg_format = '\t{}: {:.2e}'
                if level.lower() == 'info':
                    logger.info(msg_format.format(k, v))
                else:
                    logger.debug(msg_format.format(k, v))

    @staticmethod
    def early_stop(history, column, threshold=0.005, n_epoch=5):
        """Determine whether to stop training early based on nearly no change
        to validation loss for a certain number of consecutive epochs.

        Parameters
        ----------
        history : pd.DataFrame | None
            Model training history
        column : str
            Column from the model training history to evaluate for early
            termination.
        threshold : float
            The absolute relative fractional difference in validation loss
            between subsequent epochs below which an early termination is
            warranted. E.g. if val losses were 0.1 and 0.0998 the relative
            diff would be calculated as 0.0002 / 0.1 = 0.002 which would be
            less than the default thresold of 0.01 and would satisfy the
            condition for early termination.
        n_epoch : int
            The number of consecutive epochs that satisfy the threshold that
            warrants an early stop.

        Returns
        -------
        stop : bool
            Flag to stop training (True) or keep going (False).
        """
        stop = False

        if history is not None and len(history) > n_epoch + 1:
            diffs = np.abs(np.diff(history[column]))
            if all(diffs[-n_epoch:] < threshold):
                stop = True
                logger.info('Found early stop condition, loss values "{}" '
                            'have absolute relative differences less than '
                            'threshold {}: {}'
                            .format(column, threshold, diffs[-n_epoch:]))

        return stop

    def finish_epoch(self, epoch, epochs, t0, loss_details,
                     checkpoint_int, out_dir,
                     early_stop_on, early_stop_threshold,
                     early_stop_n_epoch, extras=None):
        """Perform finishing checks after an epoch is done training

        Parameters
        ----------
        epoch : int
            Epoch number that is finishing
        epochs : list
            List of epochs being iterated through
        t0 : float
            Starting time of training.
        loss_details : dict
            Namespace of the breakdown of loss components
        checkpoint_int : int | None
            Epoch interval at which to save checkpoint models.
        out_dir : str
            Directory to save checkpoint GAN models. Should have {epoch} in
            the directory name. This directory will be created if it does not
            already exist.
        early_stop_on : str | None
            If not None, this should be a column in the training history to
            evaluate for early stopping (e.g. validation_loss_gen,
            validation_loss_disc). If this value in this history decreases by
            an absolute fractional relative difference of less than 0.01 for
            more than 5 epochs in a row, the training will stop early.
        early_stop_threshold : float
            The absolute relative fractional difference in validation loss
            between subsequent epochs below which an early termination is
            warranted. E.g. if val losses were 0.1 and 0.0998 the relative
            diff would be calculated as 0.0002 / 0.1 = 0.002 which would be
            less than the default thresold of 0.01 and would satisfy the
            condition for early termination.
        early_stop_n_epoch : int
            The number of consecutive epochs that satisfy the threshold that
            warrants an early stop.
        extras : dict | None
            Extra kwargs/parameters to save in the epoch history.

        Returns
        -------
        stop : bool
            Flag to early stop training.
        """

        self.log_loss_details(loss_details)

        self._history.at[epoch, 'elapsed_time'] = time.time() - t0
        for key, value in loss_details.items():
            if key != 'n_obs':
                self._history.at[epoch, key] = value

        last_epoch = epoch == epochs[-1]
        chp = checkpoint_int is not None and (epoch % checkpoint_int) == 0
        if last_epoch or chp:
            msg = ('GAN output dir for checkpoint models should have '
                   f'{"{epoch}"} but did not: {out_dir}')
            assert '{epoch}' in out_dir, msg
            self.save(out_dir.format(epoch=epoch))

        stop = False
        if early_stop_on is not None and early_stop_on in self._history:
            stop = self.early_stop(self._history, early_stop_on,
                                   threshold=early_stop_threshold,
                                   n_epoch=early_stop_n_epoch)
            if stop:
                self.save(out_dir.format(epoch=epoch))

        if extras is not None:
            for k, v in extras.items():
                self._history.at[epoch, k] = v

        return stop

    def run_gradient_descent(self, low_res, hi_res_true, training_weights,
                             optimizer=None, **calc_loss_kwargs):
        """Run gradient descent for one mini-batch of (low_res, hi_res_true)
        and adjust NN weights

        Parameters
        ----------
        low_res : np.ndarray
            Real low-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        hi_res_true : np.ndarray
            Real high-resolution data in a 4D or 5D array:
            (n_observations, spatial_1, spatial_2, features)
            (n_observations, spatial_1, spatial_2, temporal, features)
        training_weights : list
            A list of layer weights that are to-be-trained based on the
            current loss weight values.
        optimizer : tf.keras.optimizers.Optimizer
            Optimizer class to use to update weights. This can be different if
            you're training just the generator or one of the discriminator
            models. Defaults to the generator optimizer.
        calc_loss_kwargs : dict
            Kwargs to pass to the self.calc_loss() method

        Returns
        -------
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(training_weights)

            hi_res_gen = self._tf_generate(low_res)
            loss_out = self.calc_loss(hi_res_true, hi_res_gen,
                                      **calc_loss_kwargs)
            loss, loss_details = loss_out

            grad = tape.gradient(loss, training_weights)

        if optimizer is None:
            optimizer = self.optimizer

        optimizer.apply_gradients(zip(grad, training_weights))

        return loss_details

    @tf.function
    def calc_loss_gen_content(self, hi_res_true, hi_res_gen):
        """Calculate the content loss term for the generator model.

        Parameters
        ----------
        hi_res_true : tf.Tensor
            Ground truth high resolution spatiotemporal data.
        hi_res_gen : tf.Tensor
            Superresolved high resolution spatiotemporal data generated by the
            generative model.

        Returns
        -------
        loss_gen_s : tf.Tensor
            0D tensor generator model loss for the content loss comparing the
            hi res ground truth to the hi res synthetically generated output.
        """

        loss_gen_content = self.loss_fun(hi_res_true, hi_res_gen)

        return loss_gen_content

    @staticmethod
    @tf.function
    def calc_loss_gen_advers(disc_out_gen):
        """Calculate the adversarial component of the loss term for the
        generator model.

        Parameters
        ----------
        disc_out_gen : tf.Tensor
            Raw discriminator outputs from the discriminator model
            predicting only on hi_res_gen (not on hi_res_true).

        Returns
        -------
        loss_gen_advers : tf.Tensor
            0D tensor generator model loss for the adversarial component of the
            generator loss term.
        """

        # note that these have flipped labels from the discriminator
        # loss because of the opposite optimization goal
        loss_gen_advers = tf.nn.sigmoid_cross_entropy_with_logits(
            logits=disc_out_gen, labels=tf.ones_like(disc_out_gen))
        loss_gen_advers = tf.reduce_mean(loss_gen_advers)

        return loss_gen_advers

    @staticmethod
    @tf.function
    def calc_loss_disc(disc_out_true, disc_out_gen):
        """Calculate the loss term for the discriminator model (either the
        spatial or temporal discriminator).

        Parameters
        ----------
        disc_out_true : tf.Tensor
            Raw discriminator outputs from the discriminator model predicting
            only on ground truth data hi_res_true (not on hi_res_gen).
        disc_out_gen : tf.Tensor
            Raw discriminator outputs from the discriminator model predicting
            only on synthetic data hi_res_gen (not on hi_res_true).

        Returns
        -------
        loss_disc : tf.Tensor
            0D tensor discriminator model loss for either the spatial or
            temporal component of the super resolution generated output.
        """

        # note that these have flipped labels from the generator
        # loss because of the opposite optimization goal
        logits = tf.concat([disc_out_true, disc_out_gen], axis=0)
        labels = tf.concat([tf.ones_like(disc_out_true),
                            tf.zeros_like(disc_out_gen)], axis=0)

        loss_disc = tf.nn.sigmoid_cross_entropy_with_logits(logits=logits,
                                                            labels=labels)
        loss_disc = tf.reduce_mean(loss_disc)

        return loss_disc

    @tf.function
    def calc_loss(self, hi_res_true, hi_res_gen, weight_gen_advers=0.001,
                  train_gen=True, train_disc=False):
        """Calculate the GAN loss function using generated and true high
        resolution data.

        Parameters
        ----------
        hi_res_true : tf.Tensor
            Ground truth high resolution spatiotemporal data.
        hi_res_gen : tf.Tensor
            Superresolved high resolution spatiotemporal data generated by the
            generative model.
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        train_gen : bool
            True if generator is being trained, then loss=loss_gen
        train_disc : bool
            True if disc is being trained, then loss=loss_disc

        Returns
        -------
        loss : tf.Tensor
            0D tensor representing the loss value for the network being trained
            (either generator or one of the discriminators)
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        if hi_res_gen.shape != hi_res_true.shape:
            msg = ('The tensor shapes of the synthetic output {} and '
                   'true high res {} did not have matching shape! '
                   'Check the spatiotemporal enhancement multipliers in your '
                   'your model config and data handlers.'
                   .format(hi_res_gen.shape, hi_res_true.shape))
            logger.error(msg)
            raise RuntimeError(msg)

        disc_out_true = self._tf_discriminate(hi_res_true)
        disc_out_gen = self._tf_discriminate(hi_res_gen)

        loss_gen_content = self.calc_loss_gen_content(hi_res_true, hi_res_gen)
        loss_gen_advers = self.calc_loss_gen_advers(disc_out_gen)
        loss_gen = (loss_gen_content + weight_gen_advers * loss_gen_advers)

        loss_disc = self.calc_loss_disc(disc_out_true, disc_out_gen)

        loss = None
        if train_gen:
            loss = loss_gen
        elif train_disc:
            loss = loss_disc

        loss_details = {'loss_gen': loss_gen,
                        'loss_gen_content': loss_gen_content,
                        'loss_gen_advers': loss_gen_advers,
                        'loss_disc': loss_disc,
                        }

        return loss, loss_details

    def calc_val_loss(self, batch_handler, weight_gen_advers, loss_details):
        """Calculate the validation loss at the current state of model training

        Parameters
        ----------
        batch_handler : sup3r.data_handling.preprocessing.BatchHandler
            BatchHandler object to iterate through
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        loss_details : dict
            Namespace of the breakdown of loss components

        Returns
        -------
        loss_details : dict
            Same as input but now includes val_* loss info
        """
        logger.debug('Starting end-of-epoch validation loss calculation...')
        loss_details['n_obs'] = 0
        for val_batch in batch_handler.val_data:
            high_res_gen = self._tf_generate(val_batch.low_res)
            _, v_loss_details = self.calc_loss(
                val_batch.high_res, high_res_gen,
                weight_gen_advers=weight_gen_advers,
                train_gen=False, train_disc=False)

            loss_details = self.update_loss_details(loss_details,
                                                    v_loss_details,
                                                    len(val_batch),
                                                    prefix='val_')

        return loss_details

    def train_epoch(self, batch_handler, weight_gen_advers, train_gen,
                    train_disc, disc_loss_bounds):
        """Train the GAN for one epoch.

        Parameters
        ----------
        batch_handler : sup3r.data_handling.preprocessing.BatchHandler
            BatchHandler object to iterate through
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        train_gen : bool
            Flag whether to train the generator for this set of epochs
        train_disc : bool
            Flag whether to train the discriminator for this set of epochs
        disc_loss_bounds : tuple
            Lower and upper bounds for the discriminator loss outside of which
            the discriminators will not train unless train_disc=True or
            and train_gen=False.

        Returns
        -------
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        disc_th_low = np.min(disc_loss_bounds)
        disc_th_high = np.max(disc_loss_bounds)
        loss_details = {'n_obs': 0, 'train_loss_disc': 0}

        only_gen = train_gen and not train_disc
        only_disc = train_disc and not train_gen

        for ib, batch in enumerate(batch_handler):
            trained_gen = False
            trained_disc = False
            b_loss_details = {}
            loss_disc = loss_details['train_loss_disc']
            disc_too_good = loss_disc <= disc_th_low
            disc_too_bad = (loss_disc > disc_th_high) and train_disc
            gen_too_good = disc_too_bad

            if only_gen or (train_gen and not gen_too_good):
                trained_gen = True
                b_loss_details = self.run_gradient_descent(
                    batch.low_res, batch.high_res, self.generator_weights,
                    weight_gen_advers=weight_gen_advers,
                    optimizer=self.optimizer,
                    train_gen=True, train_disc=False)

            if only_disc or (train_disc and not disc_too_good):
                trained_disc = True
                b_loss_details = self.run_gradient_descent(
                    batch.low_res, batch.high_res, self.discriminator_weights,
                    weight_gen_advers=weight_gen_advers,
                    optimizer=self.optimizer_disc,
                    train_gen=False, train_disc=True)

            b_loss_details['gen_trained_frac'] = float(trained_gen)
            b_loss_details['disc_trained_frac'] = float(trained_disc)
            loss_details = self.update_loss_details(loss_details,
                                                    b_loss_details,
                                                    len(batch),
                                                    prefix='train_')

            logger.debug('Batch {} out of {} has epoch-average '
                         '(gen / disc) loss of: ({:.2e} / {:.2e}). '
                         'Trained (gen / disc): ({} / {})'
                         .format(ib, len(batch_handler),
                                 loss_details['train_loss_gen'],
                                 loss_details['train_loss_disc'],
                                 trained_gen, trained_disc))

            if all([not trained_gen, not trained_disc]):
                msg = ('For some reason none of the GAN networks trained '
                       'during batch {} out of {}!'
                       .format(ib, len(batch_handler)))
                logger.warning(msg)
                warn(msg)

        return loss_details

    def update_adversarial_weights(self, history, adaptive_update_fraction,
                                   adaptive_update_bounds,
                                   weight_gen_advers,
                                   train_disc):
        """Update spatial / temporal adversarial loss weights based on training
        fraction history.

        Parameters
        ----------
        history : dict
            Dictionary with information on how often discriminators
            were trained during current and previous epochs.
        adaptive_update_fraction : float
            Amount by which to increase or decrease adversarial loss weights
            for adaptive updates
        adaptive_update_bounds : tuple
            Tuple specifying allowed range for history[comparison_key]. If
            history[comparison_key] < update_bounds[0] then the weight will be
            increased by (1 + update_frac). If history[comparison_key] >
            update_bounds[1] then the weight will be decreased by 1 / (1 +
            update_frac).
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        train_disc : bool
            Whether the discriminator was set to be trained during the
            previous epoch

        Returns
        -------
        weight_gen_advers : float
            Updated weight factor for the adversarial loss component of the
            generator vs. the discriminator.
        """

        if adaptive_update_fraction > 0:
            update_frac = 1
            if train_disc:
                update_frac = self.get_weight_update_fraction(
                    history, 'train_disc_trained_frac',
                    update_frac=adaptive_update_fraction,
                    update_bounds=adaptive_update_bounds)
                weight_gen_advers *= update_frac

            if update_frac != 1:
                logger.debug(
                    f'New discriminator weight: {weight_gen_advers:.3f}')

        return weight_gen_advers

    def train(self, batch_handler, n_epoch,
              weight_gen_advers=0.001,
              train_gen=True,
              train_disc=True,
              disc_loss_bounds=(0.45, 0.6),
              checkpoint_int=None,
              out_dir='./gan_{epoch}',
              early_stop_on=None,
              early_stop_threshold=0.005,
              early_stop_n_epoch=5,
              adaptive_update_bounds=(0.9, 0.99),
              adaptive_update_fraction=0.0):
        """Train the GAN model on real low res data and real high res data

        Parameters
        ----------
        batch_handler : sup3r.data_handling.preprocessing.BatchHandler
            BatchHandler object to iterate through
        n_epoch : int
            Number of epochs to train on
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        train_gen : bool
            Flag whether to train the generator for this set of epochs
        train_disc : bool
            Flag whether to train the discriminator for this set of epochs
        disc_loss_bounds : tuple
            Lower and upper bounds for the discriminator loss outside of which
            the discriminator will not train unless train_disc=True and
            train_gen=False.
        checkpoint_int : int | None
            Epoch interval at which to save checkpoint models.
        out_dir : str
            Directory to save checkpoint GAN models. Should have {epoch} in
            the directory name. This directory will be created if it does not
            already exist.
        early_stop_on : str | None
            If not None, this should be a column in the training history to
            evaluate for early stopping (e.g. validation_loss_gen,
            validation_loss_disc). If this value in this history decreases by
            an absolute fractional relative difference of less than 0.01 for
            more than 5 epochs in a row, the training will stop early.
        early_stop_threshold : float
            The absolute relative fractional difference in validation loss
            between subsequent epochs below which an early termination is
            warranted. E.g. if val losses were 0.1 and 0.0998 the relative
            diff would be calculated as 0.0002 / 0.1 = 0.002 which would be
            less than the default thresold of 0.01 and would satisfy the
            condition for early termination.
        early_stop_n_epoch : int
            The number of consecutive epochs that satisfy the threshold that
            warrants an early stop.
        adaptive_update_bounds : tuple
            Tuple specifying allowed range for loss_details[comparison_key]. If
            history[comparison_key] < threshold_range[0] then the weight will
            be increased by (1 + update_frac). If history[comparison_key] >
            threshold_range[1] then the weight will be decreased by 1 / (1 +
            update_frac).
        adaptive_update_fraction : float
            Amount by which to increase or decrease adversarial weights for
            adaptive updates
        """

        self.set_norm_stats(batch_handler.means, batch_handler.stds)
        self.set_feature_names(batch_handler.training_features,
                               batch_handler.output_features)

        epochs = list(range(n_epoch))

        if self._history is None:
            self._history = pd.DataFrame(
                columns=['elapsed_time'])
            self._history.index.name = 'epoch'
        else:
            epochs += self._history.index.values[-1] + 1

        t0 = time.time()
        logger.info('Training model with adversarial weight: {} '
                    'for {} epochs starting at epoch {}'
                    .format(weight_gen_advers, n_epoch, epochs[0]))

        for epoch in epochs:
            loss_details = self.train_epoch(batch_handler, weight_gen_advers,
                                            train_gen, train_disc,
                                            disc_loss_bounds)

            loss_details = self.calc_val_loss(batch_handler, weight_gen_advers,
                                              loss_details)

            msg = f'Epoch {epoch} of {epochs[-1]} '
            msg += 'gen/disc train loss: {:.2e}/{:.2e} '.format(
                loss_details["train_loss_gen"],
                loss_details["train_loss_disc"])

            if all(loss in loss_details for loss
                   in ('val_loss_gen', 'val_loss_disc')):
                msg += 'gen/disc val loss: {:.2e}/{:.2e} '.format(
                    loss_details["val_loss_gen"],
                    loss_details["val_loss_disc"])

            logger.info(msg)

            lr_g = self.get_optimizer_config(self.optimizer)['learning_rate']
            lr_d = self.get_optimizer_config(
                self.optimizer_disc)['learning_rate']

            extras = {'weight_gen_advers': weight_gen_advers,
                      'disc_loss_bound_0': disc_loss_bounds[0],
                      'disc_loss_bound_1': disc_loss_bounds[1],
                      'learning_rate_gen': lr_g,
                      'learning_rate_disc': lr_d}

            weight_gen_advers = self.update_adversarial_weights(
                loss_details, adaptive_update_fraction, adaptive_update_bounds,
                weight_gen_advers, train_disc)

            stop = self.finish_epoch(epoch, epochs, t0, loss_details,
                                     checkpoint_int, out_dir,
                                     early_stop_on, early_stop_threshold,
                                     early_stop_n_epoch, extras=extras)

            if stop:
                break



# -*- coding: utf-8 -*-
"""Utilities module for preparing
training data

@author: bbenton
"""

import numpy as np
import logging
from scipy import ndimage as nd
from fnmatch import fnmatch
import os
import xarray as xr
import re
from warnings import warn
import psutil
import warnings

from rex import Resource

np.random.seed(42)

logger = logging.getLogger(__name__)


def estimate_max_workers(max_workers, process_mem, n_processes):
    """Estimate max number of workers based on available memory

    Parameters
    ----------
    max_workers : int | None
        Max number of workers available
    process_mem : int
        Total number of bytes for minimum size process
    n_processes : int
        Number of processes

    Returns
    -------
    int
        Max number of workers available
    """
    if max_workers is not None:
        return max_workers
    mem = psutil.virtual_memory()
    avail_mem = 0.7 * (mem.total - mem.used)
    msg = ('Estimated upper bound for process memory '
           f'({process_mem / 1e9:.3f} GB) exceeds available memory '
           f'({avail_mem / 1e9:.3f} GB).')
    if process_mem > avail_mem:
        logger.warning(msg)
        warnings.warn(msg)
    max_workers = int(avail_mem / process_mem)
    max_workers = np.min([max_workers, n_processes, os.cpu_count()])
    max_workers = np.max([max_workers, 1])
    logger.info(f'Available memory: {avail_mem / 1e9:.3f} GB. '
                f'Max workers: {max_workers}. '
                f'Memory per process: {process_mem / 1e9:.3f} GB.')
    return max_workers


def round_array(arr, digits=3):
    """Method to round elements in an array or list. Used a lot in logging
    losses from the data-centric model

    Parameters
    ----------
    arr : list | ndarray
        List or array to round elements of
    digits : int, optional
        Number of digits to round to, by default 3

    Returns
    -------
    list
        List with rounded elements
    """
    return [round(a, digits) for a in arr]


def get_chunk_slices(arr_size, chunk_size, index_slice=slice(None)):
    """Get array slices of corresponding chunk size

    Parameters
    ----------
    arr_size : int
        Length of array to slice
    chunk_size : int
        Size of slices to split array into
    index_slice : slice
        Slice specifying starting and ending index of slice list

    Returns
    -------
    list
        List of slices corresponding to chunks of array
    """

    indices = np.arange(0, arr_size)
    indices = indices[index_slice.start:index_slice.stop]
    step = 1 if index_slice.step is None else index_slice.step
    slices = []
    start = indices[0]
    stop = start + step * chunk_size
    stop = np.min([stop, indices[-1] + 1])

    while start < indices[-1] + 1:
        slices.append(slice(start, stop, step))
        start = stop
        stop += step * chunk_size
        stop = np.min([stop, indices[-1] + 1])
    return slices


def get_file_t_steps(file_paths):
    """Get number of time steps in each file. We assume that each file in a
    file list passed to the handling classes has the same number of time steps.

    Parameters
    ----------
    file_paths : list
        List of netcdf or h5 file paths

    Returns
    -------
    int
        Number of time steps in each file
    """

    file_type = get_source_type(file_paths)

    if file_type == 'nc':
        with xr.open_dataset(file_paths[0]) as handle:
            if hasattr(handle, 'XTIME'):
                return len(handle.XTIME.values)
            elif hasattr(handle, 'time'):
                return len(handle.time.values)
    elif file_type == 'h5':
        with Resource(file_paths[0]) as handle:
            return len(handle.time_index)


def get_time_index(file_path):
    """Get time index for single file

    Parameters
    ----------
    file_path : str
        Single netcdf or h5 file

    Returns
    -------
    time_index : ndarray
        Array of time indices
    """
    file_type = get_source_type(file_path)

    if file_type == 'nc':
        with xr.open_dataset(file_path) as handle:
            if hasattr(handle, 'XTIME'):
                times = handle.XTIME.values
            elif hasattr(handle, 'time'):
                times = handle.time.values
    elif file_type == 'h5':
        with Resource(file_path) as handle:
            times = handle.time_index
    return times


def is_time_series(file_paths):
    """Check if file list is list of files with different time indices or
    different datasets

    Parameters
    ----------
    file_paths : list
        List of netcdf or h5 file paths

    Returns
    -------
    bool
        Whether files have different time indices or have the same time indices
        with different datasets
    """

    if len(file_paths) > 1:
        times_0 = get_time_index(file_paths[0])
        times_1 = get_time_index(file_paths[1])
        return not all(times_0 == times_1)
    return True


def get_raster_shape(raster_index):
    """method to get shape of raster_index"""

    if any(isinstance(r, slice) for r in raster_index):
        shape = (raster_index[0].stop - raster_index[0].start,
                 raster_index[1].stop - raster_index[1].start)
    else:
        shape = raster_index.shape
    return shape


def get_wrf_date_range(files):
    """Get wrf date range for cleaner log output. This assumes file names have
    the date pattern (YYYY-MM-DD-HH:MM:SS) or (YYYY_MM_DD_HH_MM_SS) at the end
    of the file name.

    Parameters
    ----------
    files : list
        List of wrf file paths

    Returns
    -------
    date_start : str
        start date
    date_end : str
        end date
    """

    date_start = re.search(r'(\d{4}(-|_)\d+(-|_)\d+(-|_)\d+(:|_)\d+(:|_)\d+)',
                           files[0])
    date_start = date_start if date_start is None else date_start[0]
    date_end = re.search(r'(\d{4}(-|_)\d+(-|_)\d+(-|_)\d+(:|_)\d+(:|_)\d+)',
                         files[-1])
    date_end = date_end if date_end is None else date_end[0]

    date_start = date_start.replace(':', '_')
    date_end = date_end.replace(':', '_')

    return date_start, date_end


def uniform_box_sampler(data, shape):
    '''Extracts a sample cut from data.

    Parameters
    ----------
    data : np.ndarray
        Data array with dimensions
        (spatial_1, spatial_2, temporal, features)
    shape : tuple
        (rows, cols) Size of grid to sample
        from data

    Returns
    -------
    slices : list
        List of slices corresponding to row and col extent of arr sample
    '''

    shape_1 = data.shape[0] if data.shape[0] < shape[0] else shape[0]
    shape_2 = data.shape[1] if data.shape[1] < shape[1] else shape[1]
    shape = (shape_1, shape_2)
    start_row = np.random.randint(0, data.shape[0] - shape[0] + 1)
    start_col = np.random.randint(0, data.shape[1] - shape[1] + 1)
    stop_row = start_row + shape[0]
    stop_col = start_col + shape[1]

    return [slice(start_row, stop_row), slice(start_col, stop_col)]


def weighted_time_sampler(data, shape, weights):
    """Extracts a temporal slice from data with selection weighted based on
    provided weights

    Parameters
    ----------
    data : np.ndarray
        Data array with dimensions
        (spatial_1, spatial_2, temporal, features)
    shape : tuple
        (time_steps) Size of time slice to sample
        from data
    weights : list
        List of weights used to specify selection strategy. e.g. If weights
        is [0.2, 0.8] then the start of the temporal slice will be selected
        from the first half of the temporal extent with 0.8 probability and
        0.2 probability for the second half.

    Returns
    -------
    slice : slice
        time slice with size shape
    """

    shape = data.shape[2] if data.shape[2] < shape else shape
    t_indices = (np.arange(0, data.shape[2]) if shape == 1
                 else np.arange(0, data.shape[2] - shape + 1))
    t_chunks = np.array_split(t_indices, len(weights))

    weight_list = []
    for i, w in enumerate(weights):
        weight_list += [w] * len(t_chunks[i])
    weight_list /= np.sum(weight_list)

    start = np.random.choice(t_indices, p=weight_list)
    stop = start + shape

    return slice(start, stop)


def uniform_time_sampler(data, shape):
    '''Extracts a temporal slice from data.

    Parameters
    ----------
    data : np.ndarray
        Data array with dimensions
        (spatial_1, spatial_2, temporal, features)
    shape : int
        (time_steps) Size of time slice to sample
        from data

    Returns
    -------
    slice : slice
        time slice with size shape
    '''
    shape = data.shape[2] if data.shape[2] < shape else shape
    start = np.random.randint(0, data.shape[2] - shape + 1)
    stop = start + shape
    return slice(start, stop)


def daily_time_sampler(data, shape, time_index):
    """Finds a random temporal slice from data starting at midnight

    Parameters
    ----------
    data : np.ndarray
        Data array with dimensions
        (spatial_1, spatial_2, temporal, features)
    shape : int
        (time_steps) Size of time slice to sample from data, must be an integer
        less than or equal to 24.
    time_index : pd.Datetimeindex
        Time index that matches the data axis=2

    Returns
    -------
    slice : slice
        time slice with size shape of data starting at the beginning of the day
    """

    msg = (f'data {data.shape} and time index ({len(time_index)}) '
           'shapes do not match, cannot sample daily data.')
    assert data.shape[2] == len(time_index), msg

    ti_short = time_index[:-(shape - 1)]
    midnight_ilocs = np.where((ti_short.hour == 0)
                              & (ti_short.minute == 0)
                              & (ti_short.second == 0))[0]

    if not any(midnight_ilocs):
        msg = ('Cannot sample time index of shape {} with requested daily '
               'sample shape {}'.format(len(time_index), shape))
        logger.error(msg)
        raise RuntimeError(msg)

    start = np.random.randint(0, len(midnight_ilocs))
    start = midnight_ilocs[start]
    stop = start + shape

    tslice = slice(start, stop)

    return tslice


def nsrdb_sub_daily_sampler(data, shape, time_index, csr_ind=0):
    """Finds a random sample during daylight hours of a day. Nightime is
    assumed to be marked as NaN in feature axis == csr_ind in the data input.

    Parameters
    ----------
    data : np.ndarray
        Data array with dimensions, where [..., csr_ind] is assumed to be
        clearsky ratio with NaN at night.
        (spatial_1, spatial_2, temporal, features)
    shape : int
        (time_steps) Size of time slice to sample from data, must be an integer
        less than or equal to 24.
    time_index : pd.Datetimeindex
        Time index that matches the data axis=2
    csr_ind : int
        Index of the feature axis where clearsky ratio is located and NaN's can
        be found at night.

    Returns
    -------
    tslice : slice
        time slice with size shape of data starting at the beginning of the day
    """

    tslice = daily_time_sampler(data, 24, time_index)
    night_mask = np.isnan(data[:, :, tslice, csr_ind]).any(axis=(0, 1))

    if shape == 24:
        return tslice

    if night_mask.all():
        msg = (f'No daylight data found for tslice {tslice} '
               f'{time_index[tslice]}')
        logger.warning(msg)
        warn(msg)
        return tslice

    else:
        day_ilocs = np.where(~night_mask)[0]
        padding = shape - len(day_ilocs)
        half_pad = int(np.round(padding / 2))
        new_start = tslice.start + day_ilocs[0] - half_pad
        new_end = new_start + shape
        tslice = slice(new_start, new_end)
        return tslice


def nsrdb_reduce_daily_data(data, shape, csr_ind=0):
    """Takes a 5D array and reduces the axis=3 temporal dim to daylight hours.

    Parameters
    ----------
    data : np.ndarray
        Data array 5D, where [..., csr_ind] is assumed to be
        clearsky ratio with NaN at night.
        (n_obs, spatial_1, spatial_2, temporal, features)
    shape : int
        (time_steps) Size of time slice to sample from data, must be an integer
        less than or equal to 24.
    csr_ind : int
        Index of the feature axis where clearsky ratio is located and NaN's can
        be found at night.

    Returns
    -------
    data : np.ndarray
        Same as input but with axis=3 reduced to dailylight hours with
        requested shape.
    """

    night_mask = np.isnan(data[0, :, :, :, csr_ind]).any(axis=(0, 1))

    if shape == 24:
        return data

    if night_mask.all():
        msg = (f'No daylight data found for data of shape {data.shape}')
        logger.warning(msg)
        warn(msg)
        return data

    else:
        day_ilocs = np.where(~night_mask)[0]
        padding = shape - len(day_ilocs)
        half_pad = int(np.round(padding / 2))
        start = day_ilocs[0] - half_pad
        end = start + shape
        tslice = slice(start, end)
        return data[:, :, :, tslice, :]


def transform_rotate_wind(ws, wd, lat_lon):
    """Transform windspeed/direction to u and v and align u and v with grid

    Parameters
    ----------
    ws : np.ndarray
        3D array of high res windspeed data
        (spatial_1, spatial_2, temporal)
    wd : np.ndarray
        3D array of high res winddirection data. Angle is in degrees and
        measured relative to the south_north direction.
        (spatial_1, spatial_2, temporal)
    lat_lon : np.ndarray
        3D array of lat lon
        (spatial_1, spatial_2, 2)
        Last dimension has lat / lon in that order

    Returns
    -------
    u : np.ndarray
        3D array of high res U data
        (spatial_1, spatial_2, temporal)
    v : np.ndarray
        3D array of high res V data
        (spatial_1, spatial_2, temporal)
    """
    # get the dy/dx to the nearest vertical neighbor
    dy = lat_lon[:, :, 0] - np.roll(lat_lon[:, :, 0], 1, axis=0)
    dx = lat_lon[:, :, 1] - np.roll(lat_lon[:, :, 1], 1, axis=0)

    # calculate the angle from the vertical
    theta = (np.pi / 2) - np.arctan2(dy, dx)
    del dy, dx
    theta[0] = theta[1]  # fix the roll row
    wd = np.radians(wd)

    u_rot = np.cos(theta)[:, :, np.newaxis] * ws * np.sin(wd)
    u_rot += np.sin(theta)[:, :, np.newaxis] * ws * np.cos(wd)

    v_rot = -np.sin(theta)[:, :, np.newaxis] * ws * np.sin(wd)
    v_rot += np.cos(theta)[:, :, np.newaxis] * ws * np.cos(wd)
    del theta, ws, wd
    return u_rot, v_rot


def invert_uv(u, v, lat_lon):
    """Transform u and v back to windspeed and winddirection

    Parameters
    ----------
    u : np.ndarray
        3D array of high res U data
        (spatial_1, spatial_2, temporal)
    v : np.ndarray
        3D array of high res V data
        (spatial_1, spatial_2, temporal)
    lat_lon : np.ndarray
        3D array of lat lon
        (spatial_1, spatial_2, 2)
        Last dimension has lat / lon in that order

    Returns
    -------
    ws : np.ndarray
        3D array of high res windspeed data
        (spatial_1, spatial_2, temporal)
    wd : np.ndarray
        3D array of high res winddirection data. Angle is in degrees and
        measured relative to the south_north direction.
        (spatial_1, spatial_2, temporal)
    """
    # get the dy/dx to the nearest vertical neighbor
    dy = lat_lon[:, :, 0] - np.roll(lat_lon[:, :, 0], 1, axis=0)
    dx = lat_lon[:, :, 1] - np.roll(lat_lon[:, :, 1], 1, axis=0)

    # calculate the angle from the vertical
    theta = (np.pi / 2) - np.arctan2(dy, dx)
    theta[0] = theta[1]  # fix the roll row

    u_rot = np.cos(theta)[:, :, np.newaxis] * u
    u_rot -= np.sin(theta)[:, :, np.newaxis] * v

    v_rot = np.sin(theta)[:, :, np.newaxis] * u
    v_rot += np.cos(theta)[:, :, np.newaxis] * v

    ws = np.sqrt(u_rot**2 + v_rot**2)
    wd = (np.degrees(np.arctan2(u_rot, v_rot)) + 360) % 360

    return ws, wd


def temporal_coarsening(data, t_enhance=4, method='subsample'):
    """"Coarsen data according to t_enhance resolution

    Parameters
    ----------
    data : np.ndarray
        5D array with dimensions
        (observations, spatial_1, spatial_2, temporal, features)
    t_enhance : int
        factor by which to coarsen temporal dimension
    method : str
        accepted options: [subsample, average, total]
        Subsample will take every t_enhance-th time step, average will average
        over t_enhance time steps, total will sum over t_enhance time steps

    Returns
    -------
    coarse_data : np.ndarray
        5D array with same dimensions as data with new coarse resolution
    """

    if t_enhance is not None and len(data.shape) == 5:
        if method == 'subsample':
            coarse_data = data[:, :, :, ::t_enhance, :]
        if method == 'average':
            coarse_data = np.nansum(
                data.reshape(
                    (data.shape[0], data.shape[1],
                     data.shape[2], -1, t_enhance,
                     data.shape[4])), axis=4)
            coarse_data /= t_enhance
        if method == 'total':
            coarse_data = np.nansum(
                data.reshape(
                    (data.shape[0], data.shape[1],
                     data.shape[2], -1, t_enhance,
                     data.shape[4])), axis=4)

    else:
        coarse_data = data

    return coarse_data


def daily_temporal_coarsening(data, temporal_axis=3):
    """Temporal coarsening for daily average climate change data.

    This method takes the sum of the data in the temporal dimension and divides
    by 24 (for 24 hours per day). Even if there are only 8-12 daylight obs in
    the temporal axis, we want to divide by 24 to give the equivelant of a
    daily average.

    Parameters
    ----------
    data : np.ndarray
        Array of data with a temporal axis as determined by the temporal_axis
        input. Example 4D or 5D input shapes:
        (spatial_1, spatial_2, temporal, features)
        (observations, spatial_1, spatial_2, temporal, features)
    temporal_axis : int
        Axis index of the temporal axis to be averaged. Default is axis=3 for
        the 5D tensor that is fed to the ST-GAN.

    Returns
    -------
    coarse_data : np.ndarray
        Array with same dimensions as data with new coarse resolution,
        temporal dimension is size 1
    """
    coarse_data = np.nansum(data, axis=temporal_axis) / 24
    coarse_data = np.expand_dims(coarse_data, axis=temporal_axis)
    return coarse_data


def spatial_coarsening(data, s_enhance=2, obs_axis=True):
    """"Coarsen data according to s_enhance resolution

    Parameters
    ----------
    data : np.ndarray
        3D | 4D | 5D array with dimensions:
        (n_obs, spatial_1, spatial_2, temporal, features) (obs_axis=True)
        (n_obs, spatial_1, spatial_2, features) (obs_axis=True)
        (spatial_1, spatial_2, temporal, features) (obs_axis=False)
        (spatial_1, spatial_2, features) (obs_axis=False)
    s_enhance : int
        factor by which to coarsen spatial dimensions
    obs_axis : bool
        Flag for if axis=0 is the observation axis. If True (default)
        spatial axis=(1, 2) (zero-indexed), if False spatial axis=(0, 1)

    Returns
    -------
    data : np.ndarray
        3D | 4D | 5D array with same dimensions as data with new coarse
        resolution
    """

    if len(data.shape) < 3:
        msg = ('Data must be 3D, 4D, or 5D to do spatial coarsening, but '
               f'received: {data.shape}')
        logger.error(msg)
        raise ValueError(msg)

    if s_enhance is not None and s_enhance > 1:
        bad1 = (obs_axis and (data.shape[1] % s_enhance != 0
                              or data.shape[2] % s_enhance != 0))
        bad2 = (not obs_axis and (data.shape[0] % s_enhance != 0
                                  or data.shape[1] % s_enhance != 0))
        if bad1 or bad2:
            msg = ('s_enhance must evenly divide grid size. '
                   f'Received s_enhance: {s_enhance} with data shape: '
                   f'{data.shape}')
            logger.error(msg)
            raise ValueError(msg)

        if obs_axis and len(data.shape) == 5:
            data = data.reshape(data.shape[0],
                                data.shape[1] // s_enhance, s_enhance,
                                data.shape[2] // s_enhance, s_enhance,
                                data.shape[3],
                                data.shape[4])
            data = data.sum(axis=(2, 4)) / s_enhance**2

        elif obs_axis and len(data.shape) == 4:
            data = data.reshape(data.shape[0],
                                data.shape[1] // s_enhance, s_enhance,
                                data.shape[2] // s_enhance, s_enhance,
                                data.shape[3])
            data = data.sum(axis=(2, 4)) / s_enhance**2

        elif not obs_axis and len(data.shape) == 4:
            data = data.reshape(data.shape[0] // s_enhance, s_enhance,
                                data.shape[1] // s_enhance, s_enhance,
                                data.shape[2],
                                data.shape[3])
            data = data.sum(axis=(1, 3)) / s_enhance**2

        elif not obs_axis and len(data.shape) == 3:
            data = data.reshape(data.shape[0] // s_enhance, s_enhance,
                                data.shape[1] // s_enhance, s_enhance,
                                data.shape[2])
            data = data.sum(axis=(1, 3)) / s_enhance**2

        else:
            msg = ('Data must be 3D, 4D, or 5D to do spatial coarsening, but '
                   f'received: {data.shape}')
            logger.error(msg)
            raise ValueError(msg)

    return data


def lat_lon_coarsening(lat_lon, s_enhance=2):
    """"Coarsen lat_lon according to s_enhance resolution

    Parameters
    ----------
    lat_lon : np.ndarray
        2D array with dimensions
        (spatial_1, spatial_2)
    s_enhance : int
        factor by which to coarsen spatial dimensions

    Returns
    -------
    coarse_lat_lon : np.ndarray
        2D array with same dimensions as lat_lon with new coarse resolution
    """
    coarse_lat_lon = lat_lon.reshape(-1, s_enhance,
                                     lat_lon.shape[1] // s_enhance,
                                     s_enhance, 2).sum((3, 1))
    coarse_lat_lon /= (s_enhance * s_enhance)
    return coarse_lat_lon


def forward_average(array_in):
    """Average neighboring values in an array.  Used to unstagger WRF variable
    values.

    Parameters
    ----------
    array_in : ndarray
        Input array, or array axis

    Returns
    -------
    ndarray
        Array of average values, length will be 1 less than array_in
    """
    return (array_in[:-1] + array_in[1:]) * 0.5


def unstagger_var(data, var, raster_index, time_slice=slice(None)):
    """
    Unstagger WRF variable values. Some variables use a staggered grid with
    values associated with grid cell edges. We want to center these values.

    Parameters
    ----------
    data : xarray
        netcdf data object
    var : str
        Name of variable to be unstaggered
    raster_index : list
        List of slices for raster index of spatial domain
    time_slice : slice
        slice of time to extract

    Returns
    -------
    ndarray
        Unstaggered array of variable values.
    """

    # Import Variable values from nc database instance
    array_in = np.array(data[var], np.float32)

    if all('stag' not in d for d in data[var].dims):
        array_in = array_in[(time_slice, slice(None),) + tuple(raster_index)]

    else:
        for i, d in enumerate(data[var].dims):
            if 'stag' in d:
                if 'south_north' in d:
                    idx = (time_slice, slice(None),
                           slice(raster_index[0].start,
                                 raster_index[0].stop + 1),
                           raster_index[1])
                    array_in = array_in[idx]
                elif 'west_east' in d:
                    idx = (time_slice, slice(None), raster_index[0],
                           slice(raster_index[1].start,
                                 raster_index[1].stop + 1))
                    array_in = array_in[idx]
                else:
                    idx = (time_slice, slice(None),) + tuple(raster_index)
                    array_in = array_in[idx]
                array_in = np.apply_along_axis(forward_average, i, array_in)
    return array_in


def calc_height(data, raster_index, time_slice=slice(None)):
    """Calculate height from the ground Parameters

    Parameters
    ----------
    data : xarray
        netcdf data object
    raster_index : list
        List of slices for raster index of spatial domain
    time_slice : slice
        slice of time to extract

    Returns
    -------
    height_arr : ndarray
        (temporal, vertical_level, spatial_1, spatial_2)
        4D array of heights above ground. In meters.
    """
    # Base-state Geopotential(m^2/s^2)
    if all(field in data for field in ('PHB', 'PH', 'HGT')):
        phb = unstagger_var(data, 'PHB', raster_index, time_slice)
        # Perturbation Geopotential (m^2/s^2)
        ph = unstagger_var(data, 'PH', raster_index, time_slice)
        # Terrain Height (m)
        hgt = data['HGT'][(time_slice,) + tuple(raster_index)]

        if phb.shape != hgt.shape:
            hgt = np.expand_dims(hgt, axis=1)
            hgt = np.repeat(hgt, phb.shape[-3], axis=1)
        hgt = (ph + phb) / 9.81 - hgt

    elif 'zg' in data:
        hgt = data['zg'][(time_slice, slice(None),) + tuple(raster_index)]
        hgt -= data['zg'][(time_slice, 0,) + tuple(raster_index)]
        hgt = np.array(hgt.values)

    else:
        msg = ('Need either PHB/PH/HGT or zg in data to perform height '
               'interpolation')
        raise ValueError(msg)

    return hgt


def calc_pressure(data, var, raster_index, time_slice=slice(None)):
    """Calculate pressure array

    Parameters
    ----------
    data : xarray
        netcdf data object
    var : str
        Feature to extract from data
    raster_index : list
        List of slices for raster index of spatial domain
    time_slice : slice
        slice of time to extract

    Returns
    -------
    height_arr : ndarray
        (temporal, vertical_level, spatial_1, spatial_2)
        4D array of pressure levels in pascals
    """
    p_array = np.zeros(data[var][(time_slice, slice(None),)
                                 + tuple(raster_index)].shape)
    for i in range(p_array.shape[1]):
        p_array[:, i, ...] = data.plev[i]

    return p_array


def interp_to_level(var_array, lev_array, levels):
    """Interpolate var_array to given level(s) based on h_array. Interpolation
    is linear and done for every 'z' column of [var, h] data.

    Parameters
    ----------
    var_array : ndarray
        Array of variable values
    lev_array : ndarray
        Array of height or pressure values corresponding to the wrf source data
    levels : float | list
        level or levels to interpolate to (e.g. final desired hub heights)

    Returns
    -------
    out_array : ndarray
        Array of interpolated values.
    """

    msg = ('Input arrays must be the same shape.'
           f'\nvar_array: {var_array.shape}'
           f'\nh_array: {lev_array.shape}')
    assert var_array.shape == lev_array.shape, msg

    levels = [levels] if isinstance(
        levels, (int, float, np.float32)) else levels
    h_min = np.nanmin(lev_array)
    h_max = np.nanmax(lev_array)
    height_check = (h_min <= min(levels) and max(levels) <= h_max)

    if np.isnan(h_min):
        msg = 'All pressure level height data is NaN!'
        logger.error(msg)
        raise RuntimeError(msg)

    if not height_check:
        msg = (f'Levels {levels} exceed the bounds of the pressure levels: '
               f'({h_min}, {h_max})')
        logger.warning(msg)
        warn(msg)

    array_shape = var_array.shape

    # Flatten h_array and var_array along lat, long axis
    out_array = np.zeros(
        (len(levels), array_shape[-4], np.product(array_shape[-2:]))).T

    for i in range(array_shape[0]):
        h_tmp = lev_array[i].reshape(
            (array_shape[-3], np.product(array_shape[-2:]))).T
        var_tmp = var_array[i].reshape(
            (array_shape[-3], np.product(array_shape[-2:]))).T

    # Interpolate each column of height and var to specified levels
        out_array[:, i, :] = np.array(
            [np.interp(levels, h, var)
             for h, var in zip(h_tmp, var_tmp)])

    # Reshape out_array
    if isinstance(levels, (float, np.float32, int)):
        out_array = out_array.T.reshape(
            (1, array_shape[-4], array_shape[-2], array_shape[-1]))
    else:
        out_array = out_array.T.reshape(
            (len(levels), array_shape[-4],
             array_shape[-2], array_shape[-1]))

    return out_array


def interp_var_to_height(data, var, raster_index, heights,
                         time_slice=slice(None)):
    """Interpolate var_array to given level(s) based on h_array. Interpolation
    is linear and done for every 'z' column of [var, h] data.

    Parameters
    ----------
    data : xarray
        netcdf data object
    var : str
        Name of variable to be interpolated
    raster_index : list
        List of slices for raster index of spatial domain
    heights : float | list
        level or levels to interpolate to (e.g. final desired hub heights)
    time_slice : slice
        slice of time to extract

    Returns
    -------
    out_array : ndarray
        Array of interpolated values.
    """

    logger.debug(f'Interpolating {var} to heights (m): {heights}')
    return interp_to_level(unstagger_var(data, var, raster_index, time_slice),
                           calc_height(data, raster_index, time_slice),
                           heights)[0]


def interp_var_to_pressure(data, var, raster_index, pressures,
                           time_slice=slice(None)):
    """Interpolate var_array to given level(s) based on h_array. Interpolation
    is linear and done for every 'z' column of [var, h] data.

    Parameters
    ----------
    data : xarray
        netcdf data object
    var : str
        Name of variable to be interpolated
    raster_index : list
        List of slices for raster index of spatial domain
    pressures : float | list
        level or levels to interpolate to (e.g. final desired hub heights)
    time_slice : slice
        slice of time to extract

    Returns
    -------
    out_array : ndarray
        Array of interpolated values.
    """

    logger.debug(f'Interpolating {var} to pressures (Pa): {pressures}')
    return interp_to_level(
        unstagger_var(data, var, raster_index, time_slice)[:, ::-1, ...],
        calc_pressure(data, var, raster_index, time_slice)[:, ::-1, ...],
        pressures)[0]


def potential_temperature(T, P):
    """Potential temperature of fluid at pressure P and temperature T

    Parameters
    ----------
    T : ndarray
        Temperature in celsius
    P : ndarray
        Pressure of fluid in Pa

    Returns
    -------
    ndarray
        Potential temperature
    """
    out = (T + np.float32(273.15))
    out *= (np.float32(100000) / P) ** np.float32(0.286)
    return out


def invert_pot_temp(PT, P):

    """Potential temperature of fluid at pressure P and temperature T

    Parameters
    ----------
    PT : ndarray
        Potential temperature in Kelvin
    P : ndarray
        Pressure of fluid in Pa

    Returns
    -------
    ndarray
        Temperature in celsius
    """
    out = PT / (np.float32(100000) / P) ** np.float32(0.286)
    out -= np.float32(273.15)
    return out


def potential_temperature_difference(T_top, P_top, T_bottom, P_bottom):
    """Potential temp difference calculation

    Parameters
    ---------
    T_top : ndarray
        Temperature at higher height. Used in the approximation of potential
        temperature derivative
    T_bottom : ndarray
        Temperature at lower height. Used in the approximation of potential
        temperature derivative
    P_top : ndarray
        Pressure at higher height. Used in the approximation of potential
        temperature derivative
    P_bottom : ndarray
        Pressure at lower height. Used in the approximation of potential
        temperature derivative

    Returns
    -------
    ndarray
        Difference in potential temperature between top and bottom levels
    """
    return (potential_temperature(T_top, P_top)
            - potential_temperature(T_bottom, P_bottom))


def potential_temperature_average(T_top, P_top, T_bottom, P_bottom):
    """Potential temp average calculation

    Parameters
    ---------
    T_top : ndarray
        Temperature at higher height. Used in the approximation of potential
        temperature derivative
    T_bottom : ndarray
        Temperature at lower height. Used in the approximation of potential
        temperature derivative
    P_top : ndarray
        Pressure at higher height. Used in the approximation of potential
        temperature derivative
    P_bottom : ndarray
        Pressure at lower height. Used in the approximation of potential
        temperature derivative

    Returns
    -------
    ndarray
        Average of potential temperature between top and bottom levels
    """

    return ((potential_temperature(T_top, P_top)
            + potential_temperature(T_bottom, P_bottom)) / np.float32(2.0))


def inverse_mo_length(U_star, flux_surf):
    """Inverse Monin - Obukhov Length

    Parameters
    ----------
    U_star : ndarray
        (spatial_1, spatial_2, temporal)
        Frictional wind speed
    flux_surf : ndarray
        (spatial_1, spatial_2, temporal)
        Surface heat flux

    Returns
    -------
    ndarray
        (spatial_1, spatial_2, temporal)
        Inverse Monin - Obukhov Length
    """

    denom = -U_star ** 3 * 300
    numer = (0.41 * 9.81 * flux_surf)
    return numer / denom


def bvf_squared(T_top, T_bottom, P_top, P_bottom, delta_h):
    """
    Squared Brunt Vaisala Frequency

    Parameters
    ----------
    T_top : ndarray
        Temperature at higher height. Used in the approximation of potential
        temperature derivative
    T_bottom : ndarray
        Temperature at lower height. Used in the approximation of potential
        temperature derivative
    P_top : ndarray
        Pressure at higher height. Used in the approximation of potential
        temperature derivative
    P_bottom : ndarray
        Pressure at lower height. Used in the approximation of potential
        temperature derivative
    delta_h : float
        Difference in heights between top and bottom levels

    Results
    -------
    ndarray
        Squared Brunt Vaisala Frequency
    """

    bvf2 = np.float32(9.81 / delta_h)
    bvf2 *= potential_temperature_difference(
        T_top, P_top, T_bottom, P_bottom)
    bvf2 /= potential_temperature_average(
        T_top, P_top, T_bottom, P_bottom)

    return bvf2


def gradient_richardson_number(T_top, T_bottom, P_top, P_bottom, U_top,
                               U_bottom, V_top, V_bottom, delta_h):
    """Formula for the gradient richardson number - related to the bouyant
    production or consumption of turbulence divided by the shear production of
    turbulence. Used to indicate dynamic stability

    Parameters
    ----------
    T_top : ndarray
        Temperature at higher height. Used in the approximation of potential
        temperature derivative
    T_bottom : ndarray
        Temperature at lower height. Used in the approximation of potential
        temperature derivative
    P_top : ndarray
        Pressure at higher height. Used in the approximation of potential
        temperature derivative
    P_bottom : ndarray
        Pressure at lower height. Used in the approximation of potential
        temperature derivative
    U_top : ndarray
        Zonal wind component at higher height
    U_bottom : ndarray
        Zonal wind component at lower height
    V_top : ndarray
        Meridional wind component at higher height
    V_bottom : ndarray
        Meridional wind component at lower height
    delta_h : float
        Difference in heights between top and bottom levels

    Returns
    -------
    ndarray
        Gradient Richardson Number
    """

    ws_grad = (U_top - U_bottom) ** 2
    ws_grad += (V_top - V_bottom) ** 2
    ws_grad /= delta_h ** 2
    ws_grad[ws_grad < 1e-6] = 1e-6
    Ri = bvf_squared(
        T_top, T_bottom, P_top, P_bottom, delta_h) / ws_grad
    del ws_grad
    return Ri


def nn_fill_array(array):
    """Fill any NaN values in an np.ndarray from the nearest non-nan values.

    Parameters
    ----------
    array : np.ndarray
        Input array with NaN values

    Returns
    -------
    array : np.ndarray
        Output array with NaN values filled
    """

    nan_mask = np.isnan(array)
    indices = nd.distance_transform_edt(nan_mask, return_distances=False,
                                        return_indices=True)
    array = array[tuple(indices)]
    return array


def ignore_case_path_fetch(fp):
    """Get file path which matches fp while ignoring case

    Parameters
    ----------
    fp : str
        file path

    Returns
    -------
    str
        existing file which matches fp
    """

    dirname = os.path.dirname(fp)
    basename = os.path.basename(fp)
    for file in os.listdir(dirname):
        if fnmatch(file.lower(), basename.lower()):
            return os.path.join(dirname, file)
    return None


def rotor_area(h_bottom, h_top, radius=40):
    """Area of circular section between two heights

    Parameters
    ----------
    h_bottom : float
        Lower height
    h_top : float
        Upper height
    radius : float
        Radius of rotor. Default is 40 meters

    Returns
    -------
    area : float
    """

    x_bottom = np.sqrt(radius**2 - h_bottom**2)
    x_top = np.sqrt(radius**2 - h_top**2)
    area = h_top * x_top - h_bottom * x_bottom
    area += radius**2 * np.arctan2(h_top, x_top)
    area -= radius**2 * np.arctan2(h_bottom, x_bottom)
    return area


def rotor_equiv_ws(data, heights):
    """Calculate rotor equivalent wind speed. Follows implementation in 'How
    wind speed shear and directional veer affect the power production of a
    megawatt-scale operational wind turbine. DOI:10.5194/wes-2019-86'

    Parameters
    ----------
    data : dict
        Dictionary of arrays for windspeeds/winddirections at different hub
        heights.
        Each dictionary entry has (spatial_1, spatial_2, temporal)
    heights : list
        List of heights corresponding to the windspeeds/winddirections.
        rotor is assumed to be at mean(heights).

    Returns
    -------
    rews : ndarray
        Array of rotor equivalent windspeeds.
        (spatial_1, spatial_2, temporal)
    """

    rotor_center = np.mean(heights)
    rel_heights = [h - rotor_center for h in heights]
    areas = [rotor_area(rel_heights[i], rel_heights[i + 1])
             for i in range(len(rel_heights) - 1)]
    total_area = np.sum(areas)
    areas /= total_area
    rews = np.zeros(data[list(data.keys())[0]].shape)
    for i in range(len(heights) - 1):
        ws_0 = data[f'windspeed_{heights[i]}m']
        ws_1 = data[f'windspeed_{heights[i + 1]}m']
        wd_0 = data[f'winddirection_{heights[i]}m']
        wd_1 = data[f'winddirection_{heights[i + 1]}m']
        ws_cos_0 = np.cos(np.radians(wd_0)) * ws_0
        ws_cos_1 = np.cos(np.radians(wd_1)) * ws_1
        rews += areas[i] * (ws_cos_0 + ws_cos_1)**3

    rews = 0.5 * np.cbrt(rews)
    return rews


def get_source_type(file_paths):
    """Get data source type

    Parameters
    ----------
    file_paths : list
        path to data file

    Returns
    -------
    source_type : str
        Either h5 or nc
    """
    if file_paths is None:
        return None

    if not isinstance(file_paths, list):
        file_paths = [file_paths]

    _, source_type = os.path.splitext(file_paths[0])
    if source_type == '.h5':
        return 'h5'
    else:
        return 'nc'



# -*- coding: utf-8 -*-
"""
Sup3r forward pass handling module.

@author: bbenton
"""
import numpy as np
import logging
from concurrent.futures import as_completed
from datetime import datetime as dt
import os
import warnings
import glob

from rex.utilities.execution import SpawnProcessPool
from rex.utilities.fun_utils import get_fun_call_str

from sup3r.preprocessing.data_handling import (DataHandlerH5,
                                               DataHandlerNC,
                                               DataHandlerNCforCC)
from sup3r.postprocessing.file_handling import (OutputHandlerH5,
                                                OutputHandlerNC)
from sup3r.utilities.utilities import (get_chunk_slices,
                                       get_source_type,
                                       is_time_series)
from sup3r.models import Sup3rGan

np.random.seed(42)

logger = logging.getLogger(__name__)


class ForwardPassStrategy:
    """Class to prepare data for forward passes through generator.

    A full file list of contiguous times is provided.  This file list is split
    up into chunks each of which is passed to a different node. These chunks
    can overlap in time. The file list chunk is further split up into temporal
    and spatial chunks which can also overlap. This handler stores information
    on these chunks, how they overlap, and how to crop generator output to
    stich the chunks back togerther.
    """

    def __init__(self, file_paths,
                 target=None, shape=None,
                 temporal_slice=slice(None),
                 forward_pass_chunk_shape=(100, 100, 100),
                 raster_file=None,
                 s_enhance=3,
                 t_enhance=4,
                 extract_workers=None,
                 compute_workers=None,
                 pass_workers=None,
                 temporal_extract_chunk_size=100,
                 cache_file_prefix=None,
                 out_pattern=None,
                 overwrite_cache=False,
                 spatial_overlap=15,
                 temporal_overlap=15):
        """Use these inputs to initialize data handlers on different nodes and
        to define the size of the data chunks that will be passed through the
        generator.

        Parameters
        ----------
        file_paths : list | str
            A list of files to extract raster data from. Each file must have
            the same number of timesteps. Can also pass a string with a
            unix-style file path which will be passed through glob.glob
        target : tuple
            (lat, lon) lower left corner of raster. Either need target+shape or
            raster_file.
        shape : tuple
            (rows, cols) grid size. Either need target+shape or raster_file.
        temporal_slice : slice
            Slice defining size of full temporal domain. e.g. If we have 5
            files each with 5 time steps then temporal_slice = slice(None) will
            select all 25 time steps.
        forward_pass_chunk_shape : tuple
            Max shape (spatial_1, spatial_2, temporal) of an unpadded chunk to
            use for a forward pass. The number of nodes that the
            ForwardPassStrategy is set to distribute to is calculated by
            dividing up the total time index from all file_paths by the
            temporal part of this chunk shape. Each node will then be
            parallelized accross parallel processes by the spatial chunk shape.
            If temporal_overlap / spatial_overlap are non zero the chunk sent
            to the generator can be bigger than this shape. If running in
            serial set this equal to the shape of the full spatiotemporal data
            volume for best performance.
        raster_file : str | None
            File for raster_index array for the corresponding target and
            shape. If specified the raster_index will be loaded from the file
            if it exists or written to the file if it does not yet exist.
            If None raster_index will be calculated directly. Either need
            target+shape or raster_file.
        s_enhance : int
            Factor by which to enhance spatial dimensions of low resolution
            data
        t_enhance : int
            Factor by which to enhance temporal dimension of low resolution
            data
        compute_workers : int | None
            max number of workers to use for computing features. If
            compute_workers == 1 then extraction will be serialized.
        extract_workers : int | None
            max number of workers to use for data extraction. If
            extract_workers == 1 then extraction will be serialized.
        pass_workers : int | None
            Max number of workers to use for forward passes on each node. If
            pass_workers == 1 then forward passes on chunks will be
            serialized.
        temporal_extract_chunk_size : int
            Size of chunks to split time dimension into for parallel data
            extraction. If running in serial this can be set to the size
            of the full time index for best performance.
        cache_file_prefix : str
            Prefix of path to cached feature data files
        overwrite_cache : bool
            Whether to overwrite cache files storing the computed/extracted
            feature data
        out_pattern : str
            Output file pattern. Must be of form <path>/<name>_{file_id}.<ext>.
            e.g. /tmp/sup3r_job_{file_id}.h5
            Each output file will have a unique file_id filled in and the ext
            determines the output type. If None then data will be returned in
            an array and not saved.
        spatial_overlap : int
            Size of spatial overlap between chunks passed to forward passes
            for subsequent spatial stitching
        temporal_overlap : int
            Size of temporal overlap between chunks passed to forward passes
            for subsequent temporal stitching
        output_type : str
            Either nc (netcdf) or h5. This selects the extension for output
            files and determines which output handler to use when writing the
            results of the forward passes
        """

        if isinstance(file_paths, str):
            file_paths = glob.glob(file_paths)
        self.file_paths = sorted(file_paths)
        self.input_type = get_source_type(file_paths)
        self.output_type = get_source_type(out_pattern)
        self._i = 0
        self.time_series_files = is_time_series(self.file_paths)
        self.raster_file = raster_file
        self.target = target
        self.shape = shape
        self.forward_pass_chunk_shape = forward_pass_chunk_shape
        self.temporal_slice = temporal_slice
        self.pass_workers = pass_workers
        self.extract_workers = extract_workers
        self.compute_workers = compute_workers
        self.cache_file_prefix = cache_file_prefix
        self.out_files = out_pattern
        self.temporal_extract_chunk_size = temporal_extract_chunk_size
        self.overwrite_cache = overwrite_cache
        self.t_enhance = t_enhance
        self.s_enhance = s_enhance
        self.spatial_overlap = spatial_overlap
        self.temporal_overlap = temporal_overlap

        if self.input_type == 'nc':
            self.data_handler_class = DataHandlerNC
            if not self.time_series_files:
                self.data_handler_class = DataHandlerNCforCC
        elif self.input_type == 'h5':
            self.data_handler_class = DataHandlerH5

        time_indices = self.data_handler_class.get_time_index(self.file_paths)
        self.time_indices = time_indices
        out = self.get_time_chunks(time_indices, temporal_slice,
                                   forward_pass_chunk_shape, temporal_overlap)
        self.ti_slices, self.ti_pad_slices, self.file_ids = out

        self.ti_hr_crop_slices = self.get_ti_hr_crop_slices(self.ti_slices,
                                                            self.ti_pad_slices)

        self.out_files = self.get_output_file_names(
            out_files=out_pattern, file_ids=self.file_ids)

        msg = (f'Using a larger temporal_overlap {temporal_overlap} than '
               f'temporal_chunk_size {forward_pass_chunk_shape[2]}.')
        if temporal_overlap > forward_pass_chunk_shape[2]:
            logger.warning(msg)
            warnings.warn(msg)

        msg = (f'Using a larger spatial_overlap {spatial_overlap} than '
               f'spatial_chunk_size {forward_pass_chunk_shape[:2]}.')
        if any(spatial_overlap > sc for sc in forward_pass_chunk_shape[:2]):
            logger.warning(msg)
            warnings.warn(msg)

        msg = ('Using a padded chunk size '
               f'{forward_pass_chunk_shape[2] + 2 * temporal_overlap} '
               f'larger than the full temporal domain {len(time_indices)}. '
               'Should just run without temporal chunking. ')
        if (forward_pass_chunk_shape[2] + 2 * temporal_overlap
                >= len(time_indices)):
            logger.warning(msg)
            warnings.warn(msg)

    def get_node_kwargs(self, node_index):
        """Get node specific variables given an associated index

        Parameters
        ----------
        node_index : int
            Index to select node specific variables. This index selects the
            corresponding file set, cropped_file_slice, padded_file_slice,
            and sets of padded/overlapping/cropped spatial slices for spatial
            chunks

        Returns
        -------
        kwargs : dict
            Dictionary containing the node specific variables
        """

        if node_index >= len(self.file_ids):
            msg = (f'Index is out of bounds. There are {len(self.file_ids)} '
                   f'file chunks and the index requested was {node_index}.')
            raise ValueError(msg)

        out_file = self.out_files[node_index]
        ti_pad_slice = self.ti_pad_slices[node_index]
        ti_slice = self.ti_slices[node_index]
        ti_hr_crop_slice = self.ti_hr_crop_slices[node_index]
        data_shape = (self.shape[0], self.shape[1],
                      len(self.time_indices[ti_pad_slice]))
        cache_file_prefix = (None if self.cache_file_prefix is None
                             else f'{self.cache_file_prefix}_{node_index}')

        out = self.get_spatial_chunks(data_shape=data_shape)
        lr_slices, lr_pad_slices = out[:2]
        hr_slices, hr_crop_slices = out[2:]

        chunk_shape = (lr_slices[0][0].stop - lr_slices[0][0].start,
                       lr_slices[0][1].stop - lr_slices[0][1].start,
                       data_shape[2])

        kwargs = dict(file_paths=self.file_paths,
                      out_file=out_file,
                      ti_pad_slice=ti_pad_slice,
                      ti_slice=ti_slice,
                      ti_hr_crop_slice=ti_hr_crop_slice,
                      lr_slices=lr_slices,
                      lr_pad_slices=lr_pad_slices,
                      hr_slices=hr_slices,
                      hr_crop_slices=hr_crop_slices,
                      data_shape=data_shape,
                      chunk_shape=chunk_shape,
                      cache_file_prefix=cache_file_prefix,
                      node_index=node_index)

        return kwargs

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        """Iterate over all file chunks and select node specific variables

        Returns
        -------
        kwargs : dict
            Dictionary storing kwargs for ForwardPass initialization from
            config

        Raises
        ------
        StopIteration
            Stops iteration after reaching last file chunk
        """

        if self._i < len(self.ti_slices):
            kwargs = self.get_node_kwargs(self._i)
            self._i += 1
            return kwargs
        else:
            raise StopIteration

    @property
    def nodes(self):
        """Get the number of nodes that this strategy should distribute work
        to, calculated as the source time index divided by the temporal part of
        the forward_pass_chunk_shape"""
        return len(self.ti_slices)

    def file_info_logging(self, file_paths):
        """More concise file info about data files

        Parameters
        ----------
        file_paths : list
            List of file paths

        Returns
        -------
        str
            message to append to log output that does not include a huge info
            dump of file paths
        """
        return self.data_handler_class.file_info_logging(file_paths)

    @classmethod
    def get_time_chunks(cls, time_index, temporal_slice, fp_chunk_size,
                        time_overlap):
        """Calculate the number of time chunks across the full time index

        Parameters
        ----------
        time_index : ndarray
            Array of time indices across all input files
        temporal_slice : slice
            Slice selecting range for full time index
        fp_chunk_size : tuple
            Shape of data chunks passed to generator
        time_overlap : int
            Size of temporal overlap between time chunks

        Returns
        -------
        ti_chunks : list
            List of time index slices
        ti_pad_chunks : list
            List of padded time index slices
        file_ids : list
            List of ids corresponding to start and end of time index slices.
            Used to name output files.
        """

        n_chunks = len(time_index[temporal_slice]) / fp_chunk_size[2]
        n_chunks = np.int(np.ceil(n_chunks))
        ti_chunks = np.array_split(np.arange(len(time_index[temporal_slice])),
                                   n_chunks)
        ti_pad_chunks = []
        file_ids = []
        for i, chunk in enumerate(ti_chunks):
            if len(ti_chunks) > 1:
                if i == 0:
                    tmp = np.concatenate([chunk, ti_chunks[1][:time_overlap]])
                elif i == len(ti_chunks) - 1:
                    tmp = np.concatenate([ti_chunks[-2][-time_overlap:],
                                          chunk])
                else:
                    tmp = np.concatenate([ti_chunks[i - 1][-time_overlap:],
                                          chunk,
                                          ti_chunks[i + 1][:time_overlap]])
            else:
                tmp = chunk
            ti_pad_chunks.append(tmp)
            file_ids.append(i)

        ti_chunks = [slice(chunk[0], chunk[-1] + 1) for chunk in ti_chunks]
        ti_pad_chunks = [slice(chunk[0], chunk[-1] + 1)
                         for chunk in ti_pad_chunks]
        return ti_chunks, ti_pad_chunks, file_ids

    @staticmethod
    def get_output_file_names(out_files, file_ids):
        """Get output file names for each file chunk forward pass

        Parameters
        ----------
        out_files : str
            Output file pattern. Must be of the form /path/<name>_{file_id}.ext
            Each output file will have a unique file_id filled in and the ext
            determines the output type.
        file_ids : list
            List of file ids for each output file. e.g. date range

        Returns
        -------
        list
            List of output file paths
        """

        msg = ('out_files must contain {file_id} or be None. Received '
               f'{out_files}')
        assert out_files is None or '{file_id}' in out_files, msg

        out_file_list = []
        if out_files is not None:
            dirname = os.path.dirname(out_files)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            for file_id in file_ids:
                out_file = out_files.format(file_id=file_id)
                out_file_list.append(out_file)
        else:
            out_file_list = [None] * len(file_ids)
        return out_file_list

    def get_spatial_chunks(self, data_shape=None):
        """
        Get slices for small data chunks that are passed through generator

        Parameters
        ----------

        data_shape : slice
            Size of data volume corresponding to the spatial and temporal
            extent of files in file_paths.

        Returns
        -------
        lr_slices: list
            List of slices for low res data chunks which have not been padded
        lr_pad_slices : list
            List of slices which have been padded so that high res output
            can be stitched together
        hr_slices : list
            List of slices for high res data corresponding to the
            lr_slices regions
        hr_crop_slices : list
            List of slices used for cropping generator output
            when forward passes are performed on overlapping chunks
        """

        s1_slices = get_chunk_slices(data_shape[0],
                                     self.forward_pass_chunk_shape[0])
        s2_slices = get_chunk_slices(data_shape[1],
                                     self.forward_pass_chunk_shape[1])
        t_slices = [slice(None)]

        lr_pad_slices = []
        lr_slices = []
        hr_slices = []
        hr_crop_slices = []

        for s1 in s1_slices:
            for s2 in s2_slices:
                for t in t_slices:

                    lr_slices.append((s1, s2, t, slice(None)))

                    hr_slice = self.get_hr_slices([s1, s2, t])
                    hr_slices.append(tuple(hr_slice + [slice(None)]))

                    p_slice = self.get_padded_slices([s1, s2, t],
                                                     ends=list(data_shape))
                    lr_pad_slices.append(tuple(p_slice + [slice(None)]))

                    hrc_slice = self.get_hr_cropped_slices([s1, s2, t],
                                                           p_slice)
                    hr_crop_slices.append(tuple(hrc_slice + [slice(None)]))

        return (lr_slices, lr_pad_slices, hr_slices, hr_crop_slices)

    def get_padded_slices(self, slices, ends):
        """Pad slices for data chunk overlap

        Parameters
        ----------
        slices : list
            List of unpadded slices for data chunk
            (spatial_1, spatial_2, temporal)
        ends : list
            List of max indices for spatial and temporal domains
            (spatial_1, spatial_2, temporal)

        Returns
        -------
        list
            List of padded slices
            (spatial_1, spatial_2, temporal)
        """

        pad_slices = []
        for i, s in enumerate(slices):
            start = s.start
            stop = s.stop
            if start is not None:
                if i < 2:
                    start = max(start - self.spatial_overlap, 0)
                else:
                    start = max(start - self.temporal_overlap, 0)
            if stop is not None:
                if i < 2:
                    stop = min(stop + self.spatial_overlap, ends[i])
                else:
                    stop = min(stop + self.temporal_overlap, ends[i])
            pad_slices.append(slice(start, stop))
        return pad_slices

    def get_hr_slices(self, slices):
        """Get high res slices from low res slices

        Parameters
        ----------
        slices : list
            List of low res slices

        Returns
        -------
        list
            List of high res slices
            (spatial_1, spatial_2, temporal)
        """

        hr_slices = []
        for i, s in enumerate(slices):
            start = s.start
            stop = s.stop
            if start is not None:
                if i < 2:
                    start *= self.s_enhance
                else:
                    start *= self.t_enhance
            if stop is not None:
                if i < 2:
                    stop *= self.s_enhance
                else:
                    stop *= self.t_enhance
            hr_slices.append(slice(start, stop))

        return hr_slices

    def get_ti_hr_crop_slices(self, ti_slices, ti_pad_slices):
        """Get cropped temporal slices for stitching

        Parameters
        ----------
        ti_slices : list
            List of unpadded slices for time chunks
            (temporal)
        ti_pad_slices : list
            List of padded slices for time chunks
            (temporal)

        Returns
        -------
        list
            List of cropped slices
            (temporal)
        """

        cropped_slices = []
        for _, (ps, s) in enumerate(zip(ti_pad_slices, ti_slices)):
            start = s.start
            stop = s.stop
            if start is not None:
                start = self.t_enhance * (s.start - ps.start)
            if stop is not None:
                stop = self.t_enhance * (s.stop - ps.stop)

            if start is not None and start <= 0:
                start = None
            if stop is not None and stop >= 0:
                stop = None

            cropped_slices.append(slice(start, stop))
        return cropped_slices

    def get_hr_cropped_slices(self, lr_slices, lr_pad_slices):
        """Get cropped spatial and temporal slices for stitching

        Parameters
        ----------
        lr_slices : list
            List of unpadded slices for data chunk
            (spatial_1, spatial_2, temporal)
        lr_pad_slices : list
            List of padded slices for data chunk
            (spatial_1, spatial_2, temporal)

        Returns
        -------
        list
            List of cropped slices
            (spatial_1, spatial_2, temporal)
        """

        cropped_slices = []
        for i, (ps, s) in enumerate(zip(lr_pad_slices, lr_slices)):
            start = s.start
            stop = s.stop
            if start is not None:
                if i < 2:
                    start = self.s_enhance * (s.start - ps.start)
                else:
                    start = self.t_enhance * (s.start - ps.start)
            if stop is not None:
                if i < 2:
                    stop = self.s_enhance * (s.stop - ps.stop)
                else:
                    stop = self.t_enhance * (s.stop - ps.stop)

            if start is not None and start <= 0:
                start = None
            if stop is not None and stop >= 0:
                stop = None

            cropped_slices.append(slice(start, stop))
        return cropped_slices


class ForwardPass:
    """Class to run forward passes on all chunks provided by the given
    ForwardPassStrategy. The chunks provided by the strategy are all passed
    through the GAN generator to produce high resolution output.
    """

    def __init__(self, strategy, model_path, node_index=0):
        """Initialize ForwardPass with ForwardPassStrategy. The stragegy
        provides the data chunks to run forward passes on

        Parameters
        ----------
        strategy : ForwardPassStrategy
            ForwardPassStrategy instance with information on data chunks to run
            forward passes on.
        model_path : str
            Path to Sup3rGan used to generate high resolution data
        node_index : int
            Index used to select subset of full file list on which to run
            forward passes on a single node.
        """
        self.strategy = strategy
        self.model_path = model_path
        self.features = Sup3rGan.load(model_path).training_features
        self.meta_data = Sup3rGan.load(model_path).model_params
        self.node_index = node_index

        kwargs = strategy.get_node_kwargs(node_index)

        self.file_paths = kwargs['file_paths']
        self.out_file = kwargs['out_file']
        self.ti_slice = kwargs['ti_slice']
        self.ti_pad_slice = kwargs['ti_pad_slice']
        self.ti_hr_crop_slice = kwargs['ti_hr_crop_slice']
        self.lr_slices = kwargs['lr_slices']
        self.lr_pad_slices = kwargs['lr_pad_slices']
        self.hr_slices = kwargs['hr_slices']
        self.hr_crop_slices = kwargs['hr_crop_slices']
        self.data_shape = kwargs['data_shape']
        self.chunk_shape = kwargs['chunk_shape']
        self.cache_file_prefix = kwargs['cache_file_prefix']

        self.data_handler_class = strategy.data_handler_class

        if strategy.output_type == 'nc':
            self.output_handler_class = OutputHandlerNC
        elif strategy.output_type == 'h5':
            self.output_handler_class = OutputHandlerH5

        self.data_handler = self.data_handler_class(
            self.file_paths, self.features, target=self.strategy.target,
            shape=self.strategy.shape, temporal_slice=self.ti_pad_slice,
            raster_file=self.strategy.raster_file,
            extract_workers=self.strategy.extract_workers,
            compute_workers=self.strategy.compute_workers,
            cache_file_prefix=self.cache_file_prefix,
            time_chunk_size=self.strategy.temporal_extract_chunk_size,
            overwrite_cache=self.strategy.overwrite_cache,
            val_split=0.0)

        if self.cache_file_prefix is not None:
            self.data_handler.load_cached_data()

    @staticmethod
    def forward_pass_chunk(data_chunk, crop_slices, model_path):
        """Run forward pass on smallest data chunk. Each chunk has a maximum
        shape given by self.strategy.forward_pass_chunk_shape.

        Parameters
        ----------
        data_chunk : ndarray
            Data chunk to run through model generator
            (spatial_1, spatial_2, temporal, features)
        crop_slices : list
            List of slices for extracting cropped region of interest from
            output. Output can include an extra overlapping boundary to
            facilitate stitching of chunks
        model_path : str
            Path to file for Sup3rGan used to generate high resolution
            data

        Returns
        -------
        ndarray
            High resolution data generated by GAN
        """

        model = Sup3rGan.load(model_path)
        data_chunk = np.expand_dims(data_chunk, axis=0)

        hi_res = model.generate(data_chunk)

        return hi_res[0][crop_slices]

    @classmethod
    def get_node_cmd(cls, config):
        """Get a CLI call to initialize ForwardPassStrategy and run ForwardPass
        on a single node based on an input config.

        Parameters
        ----------
        config : dict
            sup3r forward pass config with all necessary args and kwargs to
            initialize ForwardPassStrategy and run ForwardPass on a single
            node.
        """

        import_str = ('from sup3r.pipeline.forward_pass '
                      f'import ForwardPassStrategy, {cls.__name__}; '
                      'from rex import init_logger')

        fps_init_str = get_fun_call_str(ForwardPassStrategy, config)

        model_path = config['model_path']
        node_index = config['node_index']
        fwp_arg_str = f'strategy, \"{model_path}\", node_index={node_index}'
        log_file = config.get('log_file', None)
        log_level = config.get('log_level', 'INFO')
        log_arg_str = '\"sup3r\", '
        log_arg_str += f'log_file=\"{log_file}\", '
        log_arg_str += f'log_level=\"{log_level}\"'

        cmd = (f"python -c \'{import_str};\n"
               f"logger = init_logger({log_arg_str});\n"
               f"strategy = {fps_init_str};\n"
               f"fwp = {cls.__name__}({fwp_arg_str});\n"
               "fwp.run()\'\n").replace('\\', '/')

        return cmd

    def run(self):
        """ForwardPass is initialized with a file_slice_index. This index
        selects a file subset from the full file list in ForwardPassStrategy.
        This routine runs forward passes on all data chunks associated with
        this file subset.
        """
        logger.info(
            f'Starting forward passes on data shape {self.data_shape}. Using '
            f'{len(self.lr_slices)} chunks each with shape of '
            f'{self.chunk_shape}, spatial_overlap of '
            f'{self.strategy.spatial_overlap} and temporal_overlap of '
            f'{self.strategy.temporal_overlap}')

        data = np.zeros(
            (self.strategy.s_enhance * self.data_shape[0],
             self.strategy.s_enhance * self.data_shape[1],
             self.strategy.t_enhance * self.data_shape[2], 2),
            dtype=np.float32)

        if self.strategy.pass_workers == 1:
            for s_high, s_low_pad, s_high_crop in zip(self.hr_slices,
                                                      self.lr_pad_slices,
                                                      self.hr_crop_slices):

                data_chunk = self.data_handler.data[s_low_pad]
                data[s_high] = ForwardPass.forward_pass_chunk(
                    data_chunk, crop_slices=s_high_crop,
                    model_path=self.model_path)
        else:
            futures = {}
            now = dt.now()
            with SpawnProcessPool(
                    max_workers=self.strategy.pass_workers) as exe:
                for s_high, s_low_pad, s_high_crop in zip(self.hr_slices,
                                                          self.lr_pad_slices,
                                                          self.hr_crop_slices):

                    data_chunk = self.data_handler.data[s_low_pad]
                    future = exe.submit(ForwardPass.forward_pass_chunk,
                                        data_chunk=data_chunk,
                                        crop_slices=s_high_crop,
                                        model_path=self.model_path)
                    meta = {'s_high': s_high}
                    futures[future] = meta

                logger.info(f'Started forward pass for {len(self.hr_slices)} '
                            f'chunks in {dt.now() - now}.')

                for i, future in enumerate(as_completed(futures)):
                    slices = futures[future]
                    data[slices['s_high']] = future.result()
                    if i % (len(futures) // 10 + 1) == 0:
                        logger.debug(f'{i+1} out of {len(futures)} forward '
                                     'passes completed.')

        logger.info('All forward passes are complete.')
        data = data[:, :, self.ti_hr_crop_slice, :]

        if self.out_file is not None:
            logger.info(f'Saving forward pass output to {self.out_file}.')
            self.output_handler_class.write_output(
                data, self.data_handler.output_features,
                self.data_handler.lat_lon,
                self.strategy.time_indices[self.ti_slice],
                self.out_file, meta_data=self.meta_data,
                max_workers=self.data_handler.extract_workers)
        else:
            return data



"""Output handling

author : @bbenton
"""
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
import numpy as np
import xarray as xr
import pandas as pd
import logging
from scipy.interpolate import RBFInterpolator
import re
from datetime import datetime as dt
import json

from sup3r.utilities.utilities import invert_uv
from sup3r.preprocessing.feature_handling import Feature

from rex.outputs import Outputs

logger = logging.getLogger(__name__)


H5_ATTRS = {'windspeed': {'fill_value': 65535, 'scale_factor': 100.0,
                          'units': 'm s-1', 'dtype': 'uint16'},
            'winddirection': {'fill_value': 65535, 'scale_factor': 100.0,
                              'units': 'degree', 'dtype': 'int16'},
            'temperature': {'fill_value': 32767, 'scale_factor': 100.0,
                            'units': 'C', 'dtype': 'int32'},
            'pressure': {'fill_value': 65535, 'scale_factor': 0.1,
                         'units': 'Pa', 'dtype': 'uint16'},
            'bvf_mo': {'fill_value': 65535, 'scale_factor': 0.1,
                       'units': 'm s-2', 'dtype': 'uint16'},
            'bvf2': {'fill_value': 65535, 'scale_factor': 0.1,
                     'units': 's-2', 'dtype': 'int16'}}


class OutputHandler:
    """Class to handle forward pass output. This includes transforming features
    back to their original form and outputting to the correct file format.
    """

    @staticmethod
    def get_lat_lon(low_res_lat_lon, shape):
        """Get lat lon arrays for high res output file

        Parameters
        ----------
        low_res_lat_lon : ndarray
            Array of lat/lon for input data.
            (spatial_1, spatial_2, 2)
            Last dimension has ordering (lat, lon)
        shape : tuple
            (lons, lats) Shape of high res grid
        Returns
        -------
        lat_lon : ndarray
            Array of lat lons for high res output file
            (spatial_1, spatial_2, 2)
            Last dimension has ordering (lat, lon)
        """

        s_enhance = shape[0] // low_res_lat_lon.shape[0]
        old_points = np.zeros((np.product(low_res_lat_lon.shape[:-1]), 2))
        new_points = np.zeros((np.product(shape), 2))
        old_lats = low_res_lat_lon[..., 0].flatten()
        old_lons = low_res_lat_lon[..., 1].flatten()

        # This shifts the indices for the old points by the downsampling
        # fraction so that we can calculate the centers of the new points with
        # origin at zero. Obviously if the shapes are the same then there
        # should be no shift
        lat_shift = 1 - low_res_lat_lon.shape[0] / shape[0]
        lon_shift = 1 - low_res_lat_lon.shape[1] / shape[1]

        new_count = old_count = 0
        for i in range(shape[0]):
            for j in range(shape[1]):
                if i % s_enhance == 0 and j % s_enhance == 0:
                    old_points[old_count, 0] = i + lat_shift
                    old_points[old_count, 1] = j + lon_shift
                    old_count += 1
                new_points[new_count, 0] = i
                new_points[new_count, 1] = j
                new_count += 1

        new_lats = RBFInterpolator(old_points, old_lats)(new_points)
        new_lons = RBFInterpolator(old_points, old_lons)(new_points)

        return np.concatenate([new_lats.reshape(shape)[..., np.newaxis],
                               new_lons.reshape(shape)[..., np.newaxis]],
                              axis=-1)

    @staticmethod
    def get_times(low_res_times, shape):
        """Get array of times for high res NETCDF output file

        Parameters
        ----------
        low_res_times : list
            List of np.datetime64 objects for input data.
        shape : int
            Number of time steps for high res time array
        Returns
        -------
        ndarray
            Array of times for high res NETCDF output file. In hours since
            1800-01-01.
        """
        t_enhance = int(shape / len(low_res_times))
        offset = (low_res_times[1] - low_res_times[0])
        time_index = np.array([low_res_times[0] + i * offset / t_enhance
                               for i in range(shape)])
        return time_index

    @abstractmethod
    def write_output(self):
        """Write high res data with new coordinates and meta data to output"""


class OutputHandlerNC(OutputHandler):
    """OutputHandler subclass for NETCDF files"""

    # pylint: disable=W0613
    @classmethod
    def write_output(cls, data, features, low_res_lat_lon,
                     low_res_times, out_file,
                     meta_data=None, max_workers=None):
        """Write forward pass output to NETCDF file

        Parameters
        ----------
        data : ndarray
            (spatial_1, spatial_2, temporal, features)
            High resolution forward pass output
        features : list
            List of feature names corresponding to the last dimension of data
        low_res_lat_lon : ndarray
            Array of lat/lon for input data.
            (spatial_1, spatial_2, 2)
            Last dimension has ordering (lat, lon)
        low_res_times : list
            List of np.datetime64 objects for input data.
        out_file : string
            Output file path
        meta_data : dict | None
            Dictionary of meta data from model
        max_workers : int | None
            Has no effect. For compliance with H5 output handler
        """

        lat_lon = cls.get_lat_lon(low_res_lat_lon, data.shape[:2])
        times = cls.get_times(low_res_times, data.shape[-2])
        coords = {'XTIME': (['Time'], times),
                  'XLAT': (['south_north', 'east_west'], lat_lon[..., 0]),
                  'XLONG': (['south_north', 'east_west'], lat_lon[..., 1])}

        data_vars = {}
        for i, f in enumerate(features):
            data_vars[f] = (['Time', 'south_north', 'east_west'],
                            np.transpose(data[..., i], (2, 0, 1)))

        attrs = {'gan_meta': json.dumps(meta_data)}

        with xr.Dataset(data_vars=data_vars, coords=coords,
                        attrs=attrs) as ncfile:
            ncfile.to_netcdf(out_file)
        logger.info(f'Saved output of size {data.shape} to: {out_file}')

    @staticmethod
    def combine_file(files, outfile):
        """Combine all chunked output files from ForwardPass into a single file

        Parameters
        ----------
        files : list
            List of chunked output files from ForwardPass runs
        outfile : str
            Output file name for combined file
        """
        ds = xr.open_mfdataset(files, combine='nested', concat_dim='Time')
        ds.to_netcdf(outfile)
        logger.info(f'Saved combined file: {outfile}')


class OutputHandlerH5(OutputHandler):
    """Class to handle writing output to H5 file"""

    @classmethod
    def get_renamed_features(cls, features):
        """Rename features based on transformation from u/v to
        windspeed/winddirection

        Parameters
        ----------
        features : list
            List of output features

        Returns
        -------
        list
            List of renamed features u/v -> windspeed/winddirection for each
            height
        """
        heights = []
        renamed_features = features.copy()
        for f in features:
            if re.match('U_(.*?)m'.lower(), f.lower()):
                heights.append(Feature.get_height(f))

        for height in heights:
            u_idx = features.index(f'U_{height}m')
            v_idx = features.index(f'V_{height}m')

            renamed_features[u_idx] = f'windspeed_{height}m'
            renamed_features[v_idx] = f'winddirection_{height}m'

        return renamed_features

    @classmethod
    def invert_uv_features(cls, data, features, lat_lon, max_workers=None):
        """Invert U/V to windspeed and winddirection. Performed in place.

        Parameters
        ----------
        data : ndarray
            High res data from forward pass
            (spatial_1, spatial_2, temporal, features)
        features : list
            List of output features
        lat_lon : ndarray
            High res lat/lon array
            (spatial_1, spatial_2, 2)
        max_workers : int | None
            Max workers to use for inverse transform. If None the maximum
            available workers will be used.
        """

        heights = []
        for f in features:
            if re.match('U_(.*?)m'.lower(), f.lower()):
                heights.append(Feature.get_height(f))

        futures = {}
        now = dt.now()
        if max_workers == 1:
            for height in heights:
                u_idx = features.index(f'U_{height}m')
                v_idx = features.index(f'V_{height}m')
                cls.invert_uv_single_pair(data, lat_lon, u_idx, v_idx)
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as exe:
                for height in heights:
                    u_idx = features.index(f'U_{height}m')
                    v_idx = features.index(f'V_{height}m')
                    future = exe.submit(cls.invert_uv_single_pair, data,
                                        lat_lon, u_idx, v_idx)
                    futures[future] = height

                logger.info(f'Started inverse transforms on {len(heights)} '
                            f'u/v pairs in {dt.now() - now}. ')

                for i, _ in enumerate(as_completed(futures)):
                    future.result()
                    logger.debug(f'{i + 1} out of {len(futures)} inverse '
                                 'transforms completed.')

    @staticmethod
    def invert_uv_single_pair(data, lat_lon, u_idx, v_idx):
        """Perform inverse transform in place on a single u/v pair.

        Parameters
        ----------
        data : ndarray
            High res data from forward pass
            (spatial_1, spatial_2, temporal, features)
        lat_lon : ndarray
            High res lat/lon array
            (spatial_1, spatial_2, 2)
        u_idx : int
            Index in data for U component to transform
        v_idx : int
            Index in data for V component to transform
        """
        ws, wd = invert_uv(data[..., u_idx], data[..., v_idx], lat_lon)
        data[..., u_idx] = ws
        data[..., v_idx] = wd

    @classmethod
    def write_output(cls, data, features, low_res_lat_lon,
                     low_res_times, out_file,
                     meta_data=None, max_workers=None):
        """Write forward pass output to H5 file

        Parameters
        ----------
        data : ndarray
            (spatial_1, spatial_2, temporal, features)
            High resolution forward pass output
        features : list
            List of feature names corresponding to the last dimension of data
        low_res_lat_lon : ndarray
            Array of lat/lon for input data.
            (spatial_1, spatial_2, 2)
            Last dimension has ordering (lat, lon)
        low_res_times : list
            List of np.datetime64 objects for input data.
        out_file : string
            Output file path
        meta_data : dict | None
            Dictionary of meta data from model
        max_workers : int | None
            Max workers to use for inverse transform. If None the maximum
            available workers will be used.
        """
        lat_lon = cls.get_lat_lon(low_res_lat_lon, data.shape[:2])
        times = cls.get_times(low_res_times, data.shape[-2])
        freq = (times[1] - times[0]) / np.timedelta64(1, 's')
        freq = pd.tseries.offsets.DateOffset(seconds=freq)
        times = pd.date_range(times[0], times[-1], freq=freq)
        cls.invert_uv_features(data, features, lat_lon,
                               max_workers=max_workers)
        renamed_features = cls.get_renamed_features(features)
        meta = pd.DataFrame({'latitude': lat_lon[..., 0].flatten(),
                             'longitude': lat_lon[..., 1].flatten()})

        with Outputs(out_file, 'w') as fh:
            fh.meta = meta
            fh._set_time_index('time_index', times)

            for i, f in enumerate(renamed_features):
                attrs = H5_ATTRS[Feature.get_basename(f)]
                flat_data = data[..., i].reshape((-1, len(times)))
                flat_data = np.transpose(flat_data, (1, 0))
                Outputs.add_dataset(out_file, f, flat_data,
                                    dtype=attrs['dtype'], attrs=attrs)

            if meta_data is not None:
                fh.run_attrs = {'gan_meta': json.dumps(meta_data)}
        logger.info(f'Saved output of size {data.shape} to: {out_file}')



# -*- coding: utf-8 -*-
"""
Sup3r pipeline config

Created on January 25 2022

@author: bnb32
"""
from reV.config.base_analysis_config import AnalysisConfig
from reV.config.pipeline import PipelineConfig
from reV.utilities.exceptions import ConfigError


class Sup3rPipelineConfig(PipelineConfig):
    """Sup3r pipeline configuration based on reV pipeline"""

    # pylint: disable=W0201
    def _parse_dirout(self):
        """Parse pipeline steps for common dirout and unique job names."""

        dirouts = []
        names = []
        for di in self.pipeline:
            for f_config in di.values():
                config = AnalysisConfig(f_config, check_keys=False,
                                        run_preflight=False)
                dirouts.append(config.dirout)

                if 'name' in config:
                    names.append(config.name)

        if len(set(dirouts)) != 1:
            raise ConfigError('Pipeline steps must have a common output '
                              'directory but received {} different '
                              'directories.'.format(len(set(dirouts))))
        else:
            self._dirout = dirouts[0]

        if len(set(names)) != len(names):
            raise ConfigError('Pipeline steps must have a unique job names '
                              'directory but received {} duplicate names.'
                              .format(len(names) - len(set(names))))



# -*- coding: utf-8 -*-
"""Super Resolving Renewable Energy Resource Data (SUP3R)"""
import os
from sup3r.version import __version__

__author__ = """Ryan King"""
__email__ = "ryan.king@nrel.gov"

SUP3R_DIR = os.path.dirname(os.path.realpath(__file__))
CONFIG_DIR = os.path.join(SUP3R_DIR, 'configs')
TEST_DATA_DIR = os.path.join(os.path.dirname(SUP3R_DIR), 'tests', 'data')
